# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Creates a new Pigweed module."""

import abc
import argparse
import dataclasses
from dataclasses import dataclass
import datetime
import logging
import os
from pathlib import Path
import re
import sys

from typing import Any, Dict, Iterable, List, Optional, Type, Union

from pw_build import generate_modules_lists

_LOG = logging.getLogger(__name__)

_PIGWEED_LICENSE = f"""
# Copyright {datetime.datetime.now().year} The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.""".lstrip()

_PIGWEED_LICENSE_CC = _PIGWEED_LICENSE.replace('#', '//')


# TODO(frolv): Adapted from pw_protobuf. Consolidate them.
class _OutputFile:
    DEFAULT_INDENT_WIDTH = 2

    def __init__(self, file: Path, indent_width: int = DEFAULT_INDENT_WIDTH):
        self._file = file
        self._content: List[str] = []
        self._indent_width: int = indent_width
        self._indentation = 0

    def line(self, line: str = '') -> None:
        if line:
            self._content.append(' ' * self._indentation)
            self._content.append(line)
        self._content.append('\n')

    def indent(
        self,
        width: Optional[int] = None,
    ) -> '_OutputFile._IndentationContext':
        """Increases the indentation level of the output."""
        return self._IndentationContext(
            self, width if width is not None else self._indent_width
        )

    @property
    def path(self) -> Path:
        return self._file

    @property
    def content(self) -> str:
        return ''.join(self._content)

    def write(self) -> None:
        print('  create  ' + str(self._file.relative_to(Path.cwd())))
        self._file.write_text(self.content)

    class _IndentationContext:
        """Context that increases the output's indentation when it is active."""

        def __init__(self, output: '_OutputFile', width: int):
            self._output = output
            self._width: int = width

        def __enter__(self):
            self._output._indentation += self._width

        def __exit__(self, typ, value, traceback):
            self._output._indentation -= self._width


class _ModuleName:
    _MODULE_NAME_REGEX = '(^[a-zA-Z]{2,})((_[a-zA-Z0-9]+)+)$'

    def __init__(self, prefix: str, main: str) -> None:
        self._prefix = prefix
        self._main = main

    @property
    def full(self) -> str:
        return f'{self._prefix}_{self._main}'

    @property
    def prefix(self) -> str:
        return self._prefix

    @property
    def main(self) -> str:
        return self._main

    @property
    def default_namespace(self) -> str:
        return f'{self._prefix}::{self._main}'

    def upper_camel_case(self) -> str:
        return ''.join(s.capitalize() for s in self._main.split('_'))

    def __str__(self) -> str:
        return self.full

    def __repr__(self) -> str:
        return self.full

    @classmethod
    def parse(cls, name: str) -> Optional['_ModuleName']:
        match = re.fullmatch(_ModuleName._MODULE_NAME_REGEX, name)
        if not match:
            return None

        return cls(match.group(1), match.group(2)[1:])


@dataclass
class _ModuleContext:
    name: _ModuleName
    dir: Path
    root_build_files: List['_BuildFile']
    sub_build_files: List['_BuildFile']
    build_systems: List[str]
    is_upstream: bool

    def build_files(self) -> Iterable['_BuildFile']:
        yield from self.root_build_files
        yield from self.sub_build_files

    def add_docs_file(self, file: Path):
        for build_file in self.root_build_files:
            build_file.add_docs_source(str(file.relative_to(self.dir)))

    def add_cc_target(self, target: '_BuildFile.CcTarget') -> None:
        for build_file in self.root_build_files:
            build_file.add_cc_target(target)

    def add_cc_test(self, target: '_BuildFile.CcTarget') -> None:
        for build_file in self.root_build_files:
            build_file.add_cc_test(target)


class _BuildFile:
    """Abstract representation of a build file for a module."""

    @dataclass
    class Target:
        name: str

        # TODO(frolv): Shouldn't be a string list as that's build system
        # specific. Figure out a way to resolve dependencies from targets.
        deps: List[str] = dataclasses.field(default_factory=list)

    @dataclass
    class CcTarget(Target):
        sources: List[Path] = dataclasses.field(default_factory=list)
        headers: List[Path] = dataclasses.field(default_factory=list)

        def rebased_sources(self, rebase_path: Path) -> Iterable[str]:
            return (str(src.relative_to(rebase_path)) for src in self.sources)

        def rebased_headers(self, rebase_path: Path) -> Iterable[str]:
            return (str(hdr.relative_to(rebase_path)) for hdr in self.headers)

    def __init__(self, path: Path, ctx: _ModuleContext):
        self._path = path
        self._ctx = ctx

        self._docs_sources: List[str] = []
        self._cc_targets: List[_BuildFile.CcTarget] = []
        self._cc_tests: List[_BuildFile.CcTarget] = []

    @property
    def path(self) -> Path:
        return self._path

    @property
    def dir(self) -> Path:
        return self._path.parent

    def add_docs_source(self, filename: str) -> None:
        self._docs_sources.append(filename)

    def add_cc_target(self, target: CcTarget) -> None:
        self._cc_targets.append(target)

    def add_cc_test(self, target: CcTarget) -> None:
        self._cc_tests.append(target)

    def write(self) -> None:
        """Writes the contents of the build file to disk."""
        file = _OutputFile(self._path, self._indent_width())

        if self._ctx.is_upstream:
            file.line(_PIGWEED_LICENSE)
            file.line()

        self._write_preamble(file)

        for target in self._cc_targets:
            file.line()
            self._write_cc_target(file, target)

        for target in self._cc_tests:
            file.line()
            self._write_cc_test(file, target)

        if self._docs_sources:
            file.line()
            self._write_docs_target(file, self._docs_sources)

        file.write()

    @abc.abstractmethod
    def _indent_width(self) -> int:
        """Returns the default indent width for the build file's code style."""

    @abc.abstractmethod
    def _write_preamble(self, file: _OutputFile) -> None:
        """Formats"""

    @abc.abstractmethod
    def _write_cc_target(
        self,
        file: _OutputFile,
        target: '_BuildFile.CcTarget',
    ) -> None:
        """Defines a C++ library target within the build file."""

    @abc.abstractmethod
    def _write_cc_test(
        self,
        file: _OutputFile,
        target: '_BuildFile.CcTarget',
    ) -> None:
        """Defines a C++ unit test target within the build file."""

    @abc.abstractmethod
    def _write_docs_target(
        self,
        file: _OutputFile,
        docs_sources: List[str],
    ) -> None:
        """Defines a documentation target within the build file."""


# TODO(frolv): The Dict here should be Dict[str, '_GnVal'] (i.e. _GnScope),
# but mypy does not yet support recursive types:
# https://github.com/python/mypy/issues/731
_GnVal = Union[bool, int, str, List[str], Dict[str, Any]]
_GnScope = Dict[str, _GnVal]


class _GnBuildFile(_BuildFile):
    _DEFAULT_FILENAME = 'BUILD.gn'
    _INCLUDE_CONFIG_TARGET = 'public_include_path'

    def __init__(
        self,
        directory: Path,
        ctx: _ModuleContext,
        filename: str = _DEFAULT_FILENAME,
    ):
        super().__init__(directory / filename, ctx)

    def _indent_width(self) -> int:
        return 2

    def _write_preamble(self, file: _OutputFile) -> None:
        # Upstream modules always require a tests target, even if it's empty.
        has_tests = len(self._cc_tests) > 0 or self._ctx.is_upstream

        imports = []

        if self._cc_targets:
            imports.append('$dir_pw_build/target_types.gni')

        if has_tests:
            imports.append('$dir_pw_unit_test/test.gni')

        if self._docs_sources:
            imports.append('$dir_pw_docgen/docs.gni')

        file.line('import("//build_overrides/pigweed.gni")\n')
        for imp in sorted(imports):
            file.line(f'import("{imp}")')

        if self._cc_targets:
            file.line()
            _GnBuildFile._target(
                file,
                'config',
                _GnBuildFile._INCLUDE_CONFIG_TARGET,
                {
                    'include_dirs': ['public'],
                    'visibility': [':*'],
                },
            )

        if has_tests:
            file.line()
            _GnBuildFile._target(
                file,
                'pw_test_group',
                'tests',
                {
                    'tests': list(f':{test.name}' for test in self._cc_tests),
                },
            )

    def _write_cc_target(
        self,
        file: _OutputFile,
        target: _BuildFile.CcTarget,
    ) -> None:
        """Defines a GN source_set for a C++ target."""

        target_vars: _GnScope = {}

        if target.headers:
            target_vars['public_configs'] = [
                f':{_GnBuildFile._INCLUDE_CONFIG_TARGET}'
            ]
            target_vars['public'] = list(target.rebased_headers(self.dir))

        if target.sources:
            target_vars['sources'] = list(target.rebased_sources(self.dir))

        if target.deps:
            target_vars['deps'] = target.deps

        _GnBuildFile._target(file, 'pw_source_set', target.name, target_vars)

    def _write_cc_test(
        self,
        file: _OutputFile,
        target: '_BuildFile.CcTarget',
    ) -> None:
        _GnBuildFile._target(
            file,
            'pw_test',
            target.name,
            {
                'sources': list(target.rebased_sources(self.dir)),
                'deps': target.deps,
            },
        )

    def _write_docs_target(
        self,
        file: _OutputFile,
        docs_sources: List[str],
    ) -> None:
        """Defines a pw_doc_group for module documentation."""
        _GnBuildFile._target(
            file,
            'pw_doc_group',
            'docs',
            {
                'sources': docs_sources,
            },
        )

    @staticmethod
    def _target(
        file: _OutputFile,
        target_type: str,
        name: str,
        args: _GnScope,
    ) -> None:
        """Formats a GN target."""

        file.line(f'{target_type}("{name}") {{')

        with file.indent():
            _GnBuildFile._format_gn_scope(file, args)

        file.line('}')

    @staticmethod
    def _format_gn_scope(file: _OutputFile, scope: _GnScope) -> None:
        """Formats all of the variables within a GN scope to a file.

        This function does not write the enclosing braces of the outer scope to
        support use from multiple formatting contexts.
        """
        for key, val in scope.items():
            if isinstance(val, int):
                file.line(f'{key} = {val}')
                continue

            if isinstance(val, str):
                file.line(f'{key} = {_GnBuildFile._gn_string(val)}')
                continue

            if isinstance(val, bool):
                file.line(f'{key} = {str(val).lower()}')
                continue

            if isinstance(val, dict):
                file.line(f'{key} = {{')
                with file.indent():
                    _GnBuildFile._format_gn_scope(file, val)
                file.line('}')
                continue

            # Format a list of strings.
            # TODO(frolv): Lists of other types?
            assert isinstance(val, list)

            if not val:
                file.line(f'{key} = []')
                continue

            if len(val) == 1:
                file.line(f'{key} = [ {_GnBuildFile._gn_string(val[0])} ]')
                continue

            file.line(f'{key} = [')
            with file.indent():
                for string in sorted(val):
                    file.line(f'{_GnBuildFile._gn_string(string)},')
            file.line(']')

    @staticmethod
    def _gn_string(string: str) -> str:
        """Converts a Python string into a string literal within a GN file.

        Accounts for the possibility of variable interpolation within GN,
        removing quotes if unnecessary:

            "string"           ->  "string"
            "string"           ->  "string"
            "$var"             ->  var
            "$var2"            ->  var2
            "$3var"            ->  "$3var"
            "$dir_pw_foo"      ->  dir_pw_foo
            "$dir_pw_foo:bar"  ->  "$dir_pw_foo:bar"
            "$dir_pw_foo/baz"  ->  "$dir_pw_foo/baz"
            "${dir_pw_foo}"    ->  dir_pw_foo

        """

        # Check if the entire string refers to a interpolated variable.
        #
        # Simple case: '$' followed a single word, e.g. "$my_variable".
        # Note that identifiers can't start with a number.
        if re.fullmatch(r'^\$[a-zA-Z_]\w*$', string):
            return string[1:]

        # GN permits wrapping an interpolated variable in braces.
        # Check for strings of the format "${my_variable}".
        if re.fullmatch(r'^\$\{[a-zA-Z_]\w*\}$', string):
            return string[2:-1]

        return f'"{string}"'


class _BazelBuildFile(_BuildFile):
    _DEFAULT_FILENAME = 'BUILD.bazel'

    def __init__(
        self,
        directory: Path,
        ctx: _ModuleContext,
        filename: str = _DEFAULT_FILENAME,
    ):
        super().__init__(directory / filename, ctx)

    def _indent_width(self) -> int:
        return 4

    def _write_preamble(self, file: _OutputFile) -> None:
        imports = ['//pw_build:pigweed.bzl']
        if self._cc_targets:
            imports.append('pw_cc_library')

        if self._cc_tests:
            imports.append('pw_cc_test')

        file.line('load(')
        with file.indent():
            for imp in sorted(imports):
                file.line(f'"{imp}",')
        file.line(')\n')

        file.line('package(default_visibility = ["//visibility:public"])\n')
        file.line('licenses(["notice"])')

    def _write_cc_target(
        self,
        file: _OutputFile,
        target: _BuildFile.CcTarget,
    ) -> None:
        _BazelBuildFile._target(
            file,
            'pw_cc_library',
            target.name,
            {
                'srcs': list(target.rebased_sources(self.dir)),
                'hdrs': list(target.rebased_headers(self.dir)),
                'includes': ['public'],
            },
        )

    def _write_cc_test(
        self,
        file: _OutputFile,
        target: '_BuildFile.CcTarget',
    ) -> None:
        _BazelBuildFile._target(
            file,
            'pw_cc_test',
            target.name,
            {
                'srcs': list(target.rebased_sources(self.dir)),
                'deps': target.deps,
            },
        )

    def _write_docs_target(
        self,
        file: _OutputFile,
        docs_sources: List[str],
    ) -> None:
        file.line('# Bazel does not yet support building docs.')
        _BazelBuildFile._target(
            file, 'filegroup', 'docs', {'srcs': docs_sources}
        )

    @staticmethod
    def _target(
        file: _OutputFile,
        target_type: str,
        name: str,
        keys: Dict[str, List[str]],
    ) -> None:
        file.line(f'{target_type}(')

        with file.indent():
            file.line(f'name = "{name}",')

            for k, vals in keys.items():
                if len(vals) == 1:
                    file.line(f'{k} = ["{vals[0]}"],')
                    continue

                file.line(f'{k} = [')
                with file.indent():
                    for val in sorted(vals):
                        file.line(f'"{val}",')
                file.line('],')

        file.line(')')


class _CmakeBuildFile(_BuildFile):
    _DEFAULT_FILENAME = 'CMakeLists.txt'

    def __init__(
        self,
        directory: Path,
        ctx: _ModuleContext,
        filename: str = _DEFAULT_FILENAME,
    ):
        super().__init__(directory / filename, ctx)

    def _indent_width(self) -> int:
        return 2

    def _write_preamble(self, file: _OutputFile) -> None:
        file.line('include($ENV{PW_ROOT}/pw_build/pigweed.cmake)')

    def _write_cc_target(
        self,
        file: _OutputFile,
        target: _BuildFile.CcTarget,
    ) -> None:
        if target.name == self._ctx.name.full:
            target_name = target.name
        else:
            target_name = f'{self._ctx.name.full}.{target.name}'

        _CmakeBuildFile._target(
            file,
            'pw_add_module_library',
            target_name,
            {
                'sources': list(target.rebased_sources(self.dir)),
                'headers': list(target.rebased_headers(self.dir)),
                'public_includes': ['public'],
            },
        )

    def _write_cc_test(
        self,
        file: _OutputFile,
        target: '_BuildFile.CcTarget',
    ) -> None:
        _CmakeBuildFile._target(
            file,
            'pw_auto_add_module_tests',
            self._ctx.name.full,
            {'private_deps': []},
        )

    def _write_docs_target(
        self,
        file: _OutputFile,
        docs_sources: List[str],
    ) -> None:
        file.line('# CMake does not yet support building docs.')

    @staticmethod
    def _target(
        file: _OutputFile,
        target_type: str,
        name: str,
        keys: Dict[str, List[str]],
    ) -> None:
        file.line(f'{target_type}({name}')

        with file.indent():
            for k, vals in keys.items():
                file.line(k.upper())
                with file.indent():
                    for val in sorted(vals):
                        file.line(val)

        file.line(')')


class _LanguageGenerator:
    """Generates files for a programming language in a new Pigweed module."""

    def __init__(self, ctx: _ModuleContext) -> None:
        self._ctx = ctx

    @abc.abstractmethod
    def create_source_files(self) -> None:
        """Creates the boilerplate source files required by the language."""


class _CcLanguageGenerator(_LanguageGenerator):
    """Generates boilerplate source files for a C++ module."""

    def __init__(self, ctx: _ModuleContext) -> None:
        super().__init__(ctx)

        self._public_dir = ctx.dir / 'public'
        self._headers_dir = self._public_dir / ctx.name.full

    def create_source_files(self) -> None:
        self._headers_dir.mkdir(parents=True)

        main_header = self._new_header(self._ctx.name.main)
        main_source = self._new_source(self._ctx.name.main)
        test_source = self._new_source(f'{self._ctx.name.main}_test')

        # TODO(frolv): This could be configurable.
        namespace = self._ctx.name.default_namespace

        main_source.line(
            f'#include "{main_header.path.relative_to(self._public_dir)}"\n'
        )
        main_source.line(f'namespace {namespace} {{\n')
        main_source.line('int magic = 42;\n')
        main_source.line(f'}}  // namespace {namespace}')

        main_header.line(f'namespace {namespace} {{\n')
        main_header.line('extern int magic;\n')
        main_header.line(f'}}  // namespace {namespace}')

        test_source.line(
            f'#include "{main_header.path.relative_to(self._public_dir)}"\n'
        )
        test_source.line('#include "gtest/gtest.h"\n')
        test_source.line(f'namespace {namespace} {{')
        test_source.line('namespace {\n')
        test_source.line(
            f'TEST({self._ctx.name.upper_camel_case()}, GeneratesCorrectly) {{'
        )
        with test_source.indent():
            test_source.line('EXPECT_EQ(magic, 42);')
        test_source.line('}\n')
        test_source.line('}  // namespace')
        test_source.line(f'}}  // namespace {namespace}')

        self._ctx.add_cc_target(
            _BuildFile.CcTarget(
                name=self._ctx.name.full,
                sources=[main_source.path],
                headers=[main_header.path],
            )
        )

        self._ctx.add_cc_test(
            _BuildFile.CcTarget(
                name=f'{self._ctx.name.main}_test',
                deps=[f':{self._ctx.name.full}'],
                sources=[test_source.path],
            )
        )

        main_header.write()
        main_source.write()
        test_source.write()

    def _new_source(self, name: str) -> _OutputFile:
        file = _OutputFile(self._ctx.dir / f'{name}.cc')

        if self._ctx.is_upstream:
            file.line(_PIGWEED_LICENSE_CC)
            file.line()

        return file

    def _new_header(self, name: str) -> _OutputFile:
        file = _OutputFile(self._headers_dir / f'{name}.h')

        if self._ctx.is_upstream:
            file.line(_PIGWEED_LICENSE_CC)

        file.line('#pragma once\n')
        return file


_BUILD_FILES: Dict[str, Type[_BuildFile]] = {
    'bazel': _BazelBuildFile,
    'cmake': _CmakeBuildFile,
    'gn': _GnBuildFile,
}

_LANGUAGE_GENERATORS: Dict[str, Type[_LanguageGenerator]] = {
    'cc': _CcLanguageGenerator,
}


def _check_module_name(
    module: str,
    is_upstream: bool,
) -> Optional[_ModuleName]:
    """Checks whether a module name is valid."""

    name = _ModuleName.parse(module)
    if not name:
        _LOG.error(
            '"%s" does not conform to the Pigweed module name format', module
        )
        return None

    if is_upstream and name.prefix != 'pw':
        _LOG.error('Modules within Pigweed itself must start with "pw_"')
        return None

    return name


def _create_main_docs_file(ctx: _ModuleContext) -> None:
    """Populates the top-level docs.rst file within a new module."""

    docs_file = _OutputFile(ctx.dir / 'docs.rst')
    docs_file.line(f'.. _module-{ctx.name}:\n')

    title = '=' * len(ctx.name.full)
    docs_file.line(title)
    docs_file.line(ctx.name.full)
    docs_file.line(title)
    docs_file.line(f'This is the main documentation file for {ctx.name}.')

    ctx.add_docs_file(docs_file.path)

    docs_file.write()


def _basic_module_setup(
    module_name: _ModuleName,
    module_dir: Path,
    build_systems: Iterable[str],
    is_upstream: bool,
) -> _ModuleContext:
    """Creates the basic layout of a Pigweed module."""
    module_dir.mkdir()

    ctx = _ModuleContext(
        name=module_name,
        dir=module_dir,
        root_build_files=[],
        sub_build_files=[],
        build_systems=list(build_systems),
        is_upstream=is_upstream,
    )

    ctx.root_build_files.extend(
        _BUILD_FILES[build](module_dir, ctx) for build in ctx.build_systems
    )

    _create_main_docs_file(ctx)

    return ctx


def _create_module(
    module: str, languages: Iterable[str], build_systems: Iterable[str]
) -> None:
    project_root = Path(os.environ.get('PW_PROJECT_ROOT', ''))
    assert project_root.is_dir()

    is_upstream = os.environ.get('PW_ROOT') == str(project_root)

    module_name = _check_module_name(module, is_upstream)
    if not module_name:
        sys.exit(1)

    if not is_upstream:
        _LOG.error(
            '`pw module create` is experimental and does '
            'not yet support downstream projects.'
        )
        sys.exit(1)

    module_dir = project_root / module

    if module_dir.is_dir():
        _LOG.error('Module %s already exists', module)
        sys.exit(1)

    if module_dir.is_file():
        _LOG.error(
            'Cannot create module %s as a file of that name already exists',
            module,
        )
        sys.exit(1)

    ctx = _basic_module_setup(
        module_name, module_dir, build_systems, is_upstream
    )

    try:
        generators = list(_LANGUAGE_GENERATORS[lang](ctx) for lang in languages)
    except KeyError as key:
        _LOG.error('Unsupported language: %s', key)
        sys.exit(1)

    for generator in generators:
        generator.create_source_files()

    for build_file in ctx.build_files():
        build_file.write()

    if is_upstream:
        modules_file = project_root / 'PIGWEED_MODULES'
        if not modules_file.exists():
            _LOG.error(
                'Could not locate PIGWEED_MODULES file; '
                'your repository may be in a bad state.'
            )
            return

        modules_gni_file = (
            project_root / 'pw_build' / 'generated_pigweed_modules_lists.gni'
        )

        # Cut off the extra newline at the end of the file.
        modules_list = modules_file.read_text().split('\n')[:-1]
        modules_list.append(module_name.full)
        modules_list.sort()
        modules_list.append('')
        modules_file.write_text('\n'.join(modules_list))
        print('  modify  ' + str(modules_file.relative_to(Path.cwd())))

        generate_modules_lists.main(
            root=project_root,
            modules_list=modules_file,
            modules_gni_file=modules_gni_file,
            mode=generate_modules_lists.Mode.UPDATE,
        )
        print('  modify  ' + str(modules_gni_file.relative_to(Path.cwd())))

    print()
    _LOG.info(
        'Module %s created at %s',
        module_name,
        module_dir.relative_to(Path.cwd()),
    )


def register_subcommand(parser: argparse.ArgumentParser) -> None:
    csv = lambda s: s.split(',')

    parser.add_argument(
        '--build-systems',
        help=(
            'Comma-separated list of build systems the module supports. '
            f'Options: {", ".join(_BUILD_FILES.keys())}'
        ),
        type=csv,
        default=_BUILD_FILES.keys(),
        metavar='BUILD[,BUILD,...]',
    )
    parser.add_argument(
        '--languages',
        help=(
            'Comma-separated list of languages the module will use. '
            f'Options: {", ".join(_LANGUAGE_GENERATORS.keys())}'
        ),
        type=csv,
        default=[],
        metavar='LANG[,LANG,...]',
    )
    parser.add_argument(
        'module', help='Name of the module to create.', metavar='MODULE_NAME'
    )
    parser.set_defaults(func=_create_module)



#!/usr/bin/env python3
# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Creates and manages token databases.

This module manages reading tokenized strings from ELF files and building and
maintaining token databases.
"""

import argparse
from datetime import datetime
import glob
import itertools
import json
import logging
import os
from pathlib import Path
import re
import struct
import sys
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Pattern,
    Set,
    TextIO,
    Tuple,
    Union,
)

try:
    from pw_tokenizer import elf_reader, tokens
except ImportError:
    # Append this path to the module search path to allow running this module
    # without installing the pw_tokenizer package.
    sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    from pw_tokenizer import elf_reader, tokens

_LOG = logging.getLogger('pw_tokenizer')


def _elf_reader(elf) -> elf_reader.Elf:
    return elf if isinstance(elf, elf_reader.Elf) else elf_reader.Elf(elf)


# Magic number used to indicate the beginning of a tokenized string entry. This
# value MUST match the value of _PW_TOKENIZER_ENTRY_MAGIC in
# pw_tokenizer/public/pw_tokenizer/internal/tokenize_string.h.
_TOKENIZED_ENTRY_MAGIC = 0xBAA98DEE
_ENTRY = struct.Struct('<4I')
_TOKENIZED_ENTRY_SECTIONS = re.compile(r'^\.pw_tokenizer.entries(?:\.[_\d]+)?$')

_ERROR_HANDLER = 'surrogateescape'  # How to deal with UTF-8 decoding errors


class Error(Exception):
    """Failed to extract token entries from an ELF file."""


def _read_tokenized_entries(
    data: bytes, domain: Pattern[str]
) -> Iterator[tokens.TokenizedStringEntry]:
    index = 0

    while index + _ENTRY.size <= len(data):
        magic, token, domain_len, string_len = _ENTRY.unpack_from(data, index)

        if magic != _TOKENIZED_ENTRY_MAGIC:
            raise Error(
                f'Expected magic number 0x{_TOKENIZED_ENTRY_MAGIC:08x}, '
                f'found 0x{magic:08x}'
            )

        start = index + _ENTRY.size
        index = start + domain_len + string_len

        # Create the entries, trimming null terminators.
        entry = tokens.TokenizedStringEntry(
            token,
            data[start + domain_len : index - 1].decode(errors=_ERROR_HANDLER),
            data[start : start + domain_len - 1].decode(errors=_ERROR_HANDLER),
        )

        if data[start + domain_len - 1] != 0:
            raise Error(
                f'Domain {entry.domain} for {entry.string} not null terminated'
            )

        if data[index - 1] != 0:
            raise Error(f'String {entry.string} is not null terminated')

        if domain.fullmatch(entry.domain):
            yield entry


def _database_from_elf(elf, domain: Pattern[str]) -> tokens.Database:
    """Reads the tokenized strings from an elf_reader.Elf or ELF file object."""
    _LOG.debug('Reading tokenized strings in domain "%s" from %s', domain, elf)

    reader = _elf_reader(elf)

    # Read tokenized string entries.
    section_data = reader.dump_section_contents(_TOKENIZED_ENTRY_SECTIONS)
    if section_data is not None:
        return tokens.Database(_read_tokenized_entries(section_data, domain))

    return tokens.Database([])


def tokenization_domains(elf) -> Iterator[str]:
    """Lists all tokenization domains in an ELF file."""
    reader = _elf_reader(elf)
    section_data = reader.dump_section_contents(_TOKENIZED_ENTRY_SECTIONS)
    if section_data is not None:
        yield from frozenset(
            e.domain
            for e in _read_tokenized_entries(section_data, re.compile('.*'))
        )


def read_tokenizer_metadata(elf) -> Dict[str, int]:
    """Reads the metadata entries from an ELF."""
    sections = _elf_reader(elf).dump_section_contents(r'\.pw_tokenizer\.info')

    metadata: Dict[str, int] = {}
    if sections is not None:
        for key, value in struct.iter_unpack('12sI', sections):
            try:
                metadata[key.rstrip(b'\0').decode()] = value
            except UnicodeDecodeError as err:
                _LOG.error(
                    'Failed to decode metadata key %r: %s',
                    key.rstrip(b'\0'),
                    err,
                )

    return metadata


def _database_from_strings(strings: List[str]) -> tokens.Database:
    """Generates a C and C++ compatible database from untokenized strings."""
    # Generate a C-compatible database from the fixed length hash.
    c_db = tokens.Database.from_strings(strings, tokenize=tokens.c_hash)

    # Generate a C++ compatible database by allowing the hash to follow the
    # string length.
    cpp_db = tokens.Database.from_strings(
        strings, tokenize=tokens.pw_tokenizer_65599_hash
    )

    # Use a union of the C and C++ compatible databases.
    return tokens.Database.merged(c_db, cpp_db)


def _database_from_json(fd) -> tokens.Database:
    return _database_from_strings(json.load(fd))


def _load_token_database(  # pylint: disable=too-many-return-statements
    db, domain: Pattern[str]
) -> tokens.Database:
    """Loads a Database from supported database types.

    Supports Database objects, JSONs, ELFs, CSVs, and binary databases.
    """
    if db is None:
        return tokens.Database()

    if isinstance(db, tokens.Database):
        return db

    if isinstance(db, elf_reader.Elf):
        return _database_from_elf(db, domain)

    # If it's a str, it might be a path. Check if it's an ELF, CSV, or JSON.
    if isinstance(db, (str, Path)):
        if not os.path.exists(db):
            raise FileNotFoundError(f'"{db}" is not a path to a token database')

        if Path(db).is_dir():
            return tokens.DatabaseFile.load(Path(db))

        # Read the path as an ELF file.
        with open(db, 'rb') as fd:
            if elf_reader.compatible_file(fd):
                return _database_from_elf(fd, domain)

        # Generate a database from JSON.
        if str(db).endswith('.json'):
            with open(db, 'r', encoding='utf-8') as json_fd:
                return _database_from_json(json_fd)

        # Read the path as a packed binary or CSV file.
        return tokens.DatabaseFile.load(Path(db))

    # Assume that it's a file object and check if it's an ELF.
    if elf_reader.compatible_file(db):
        return _database_from_elf(db, domain)

    # Read the database as JSON, CSV, or packed binary from a file object's
    # path.
    if hasattr(db, 'name') and os.path.exists(db.name):
        if db.name.endswith('.json'):
            return _database_from_json(db)

        return tokens.DatabaseFile.load(Path(db.name))

    # Read CSV directly from the file object.
    return tokens.Database(tokens.parse_csv(db))


def load_token_database(
    *databases, domain: Union[str, Pattern[str]] = tokens.DEFAULT_DOMAIN
) -> tokens.Database:
    """Loads a Database from supported database types.

    Supports Database objects, JSONs, ELFs, CSVs, and binary databases.
    """
    domain = re.compile(domain)
    return tokens.Database.merged(
        *(_load_token_database(db, domain) for db in databases)
    )


def database_summary(db: tokens.Database) -> Dict[str, Any]:
    """Returns a simple report of properties of the database."""
    present = [entry for entry in db.entries() if not entry.date_removed]
    collisions = {
        token: list(e.string for e in entries)
        for token, entries in db.collisions()
    }

    # Add 1 to each string's size to account for the null terminator.
    return dict(
        present_entries=len(present),
        present_size_bytes=sum(len(entry.string) + 1 for entry in present),
        total_entries=len(db.entries()),
        total_size_bytes=sum(len(entry.string) + 1 for entry in db.entries()),
        collisions=collisions,
    )


_DatabaseReport = Dict[str, Dict[str, Dict[str, Any]]]


def generate_reports(paths: Iterable[Path]) -> _DatabaseReport:
    """Returns a dictionary with information about the provided databases."""
    reports: _DatabaseReport = {}

    for path in paths:
        domains = ['']
        if path.is_file():
            with path.open('rb') as file:
                if elf_reader.compatible_file(file):
                    domains = list(tokenization_domains(file))

        domain_reports = {}

        for domain in domains:
            domain_reports[domain] = database_summary(
                load_token_database(path, domain=domain)
            )

        reports[str(path)] = domain_reports

    return reports


def _handle_create(
    databases,
    database: Path,
    force: bool,
    output_type: str,
    include: list,
    exclude: list,
    replace: list,
) -> None:
    """Creates a token database file from one or more ELF files."""
    if not force and database.exists():
        raise FileExistsError(
            f'The file {database} already exists! Use --force to overwrite.'
        )

    if output_type == 'directory':
        if str(database) == '-':
            raise ValueError(
                'Cannot specify "-" (stdout) for directory databases'
            )

        database.mkdir(exist_ok=True)
        database = database / f'database{tokens.DIR_DB_SUFFIX}'
        output_type = 'csv'

    if str(database) == '-':
        # Must write bytes to stdout; use sys.stdout.buffer.
        fd = sys.stdout.buffer
    else:
        fd = database.open('wb')

    db = tokens.Database.merged(*databases)
    db.filter(include, exclude, replace)

    with fd:
        if output_type == 'csv':
            tokens.write_csv(db, fd)
        elif output_type == 'binary':
            tokens.write_binary(db, fd)
        else:
            raise ValueError(f'Unknown database type "{output_type}"')

    _LOG.info(
        'Wrote database with %d entries to %s as %s',
        len(db),
        fd.name,
        output_type,
    )


def _handle_add(
    token_database: tokens.DatabaseFile,
    databases: List[tokens.Database],
    commit: Optional[str],
) -> None:
    initial = len(token_database)
    if commit:
        entries = itertools.chain.from_iterable(
            db.entries() for db in databases
        )
        token_database.add_and_discard_temporary(entries, commit)
    else:
        for source in databases:
            token_database.add(source.entries())

        token_database.write_to_file()

    number_of_changes = len(token_database) - initial

    if number_of_changes:
        _LOG.info(
            'Added %d entries to %s', number_of_changes, token_database.path
        )


def _handle_mark_removed(
    token_database: tokens.DatabaseFile,
    databases: List[tokens.Database],
    date: Optional[datetime],
):
    marked_removed = token_database.mark_removed(
        (
            entry
            for entry in tokens.Database.merged(*databases).entries()
            if not entry.date_removed
        ),
        date,
    )

    token_database.write_to_file(rewrite=True)

    _LOG.info(
        'Marked %d of %d entries as removed in %s',
        len(marked_removed),
        len(token_database),
        token_database.path,
    )


def _handle_purge(
    token_database: tokens.DatabaseFile, before: Optional[datetime]
):
    purged = token_database.purge(before)
    token_database.write_to_file(rewrite=True)

    _LOG.info('Removed %d entries from %s', len(purged), token_database.path)


def _handle_report(token_database_or_elf: List[Path], output: TextIO) -> None:
    json.dump(generate_reports(token_database_or_elf), output, indent=2)
    output.write('\n')


def expand_paths_or_globs(*paths_or_globs: str) -> Iterable[Path]:
    """Expands any globs in a list of paths; raises FileNotFoundError."""
    for path_or_glob in paths_or_globs:
        if os.path.exists(path_or_glob):
            # This is a valid path; yield it without evaluating it as a glob.
            yield Path(path_or_glob)
        else:
            paths = glob.glob(path_or_glob, recursive=True)

            # If no paths were found and the path is not a glob, raise an Error.
            if not paths and not any(c in path_or_glob for c in '*?[]!'):
                raise FileNotFoundError(f'{path_or_glob} is not a valid path')

            for path in paths:
                # Resolve globs to JSON, CSV, or compatible binary files.
                if elf_reader.compatible_file(path) or path.endswith(
                    ('.csv', '.json')
                ):
                    yield Path(path)


class ExpandGlobs(argparse.Action):
    """Argparse action that expands and appends paths."""

    def __call__(self, parser, namespace, values, unused_option_string=None):
        setattr(namespace, self.dest, list(expand_paths_or_globs(*values)))


def _read_elf_with_domain(
    elf: str, domain: Pattern[str]
) -> Iterable[tokens.Database]:
    for path in expand_paths_or_globs(elf):
        with path.open('rb') as file:
            if not elf_reader.compatible_file(file):
                raise ValueError(
                    f'{elf} is not an ELF file, '
                    f'but the "{domain}" domain was specified'
                )

            yield _database_from_elf(file, domain)


class LoadTokenDatabases(argparse.Action):
    """Argparse action that reads tokenize databases from paths or globs.

    ELF files may have #domain appended to them to specify a tokenization domain
    other than the default.
    """

    def __call__(self, parser, namespace, values, option_string=None):
        databases: List[tokens.Database] = []
        paths: Set[Path] = set()

        try:
            for value in values:
                if value.count('#') == 1:
                    path, domain = value.split('#')
                    domain = re.compile(domain)
                    databases.extend(_read_elf_with_domain(path, domain))
                else:
                    paths.update(expand_paths_or_globs(value))

            for path in paths:
                databases.append(load_token_database(path))
        except tokens.DatabaseFormatError as err:
            parser.error(
                f'argument elf_or_token_database: {path} is not a supported '
                'token database file. Only ELF files or token databases (CSV '
                f'or binary format) are supported. {err}. '
            )
        except FileNotFoundError as err:
            parser.error(f'argument elf_or_token_database: {err}')
        except:  # pylint: disable=bare-except
            _LOG.exception('Failed to load token database %s', path)
            parser.error(
                'argument elf_or_token_database: '
                f'Error occurred while loading token database {path}'
            )

        setattr(namespace, self.dest, databases)


def token_databases_parser(nargs: str = '+') -> argparse.ArgumentParser:
    """Returns an argument parser for reading token databases.

    These arguments can be added to another parser using the parents arg.
    """
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        'databases',
        metavar='elf_or_token_database',
        nargs=nargs,
        action=LoadTokenDatabases,
        help=(
            'ELF or token database files from which to read strings and '
            'tokens. For ELF files, the tokenization domain to read from '
            'may specified after the path as #domain_name (e.g. '
            'foo.elf#TEST_DOMAIN). Unless specified, only the default '
            'domain ("") is read from ELF files; .* reads all domains. '
            'Globs are expanded to compatible database files.'
        ),
    )
    return parser


def _parse_args():
    """Parse and return command line arguments."""

    def year_month_day(value) -> datetime:
        if value == 'today':
            return datetime.now()

        return datetime.fromisoformat(value)

    year_month_day.__name__ = 'year-month-day (YYYY-MM-DD)'

    # Shared command line options.
    option_db = argparse.ArgumentParser(add_help=False)
    option_db.add_argument(
        '-d',
        '--database',
        dest='token_database',
        type=lambda arg: tokens.DatabaseFile.load(Path(arg)),
        required=True,
        help='The database file to update.',
    )

    option_tokens = token_databases_parser('*')

    # Top-level argument parser.
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.set_defaults(handler=lambda **_: parser.print_help())

    subparsers = parser.add_subparsers(
        help='Tokenized string database management actions:'
    )

    # The 'create' command creates a database file.
    subparser = subparsers.add_parser(
        'create',
        parents=[option_tokens],
        help=(
            'Creates a database with tokenized strings from one or more '
            'sources.'
        ),
    )
    subparser.set_defaults(handler=_handle_create)
    subparser.add_argument(
        '-d',
        '--database',
        required=True,
        type=Path,
        help='Path to the database file to create; use - for stdout.',
    )
    subparser.add_argument(
        '-t',
        '--type',
        dest='output_type',
        choices=('csv', 'binary', 'directory'),
        default='csv',
        help='Which type of database to create. (default: csv)',
    )
    subparser.add_argument(
        '-f',
        '--force',
        action='store_true',
        help='Overwrite the database if it exists.',
    )
    subparser.add_argument(
        '-i',
        '--include',
        type=re.compile,
        default=[],
        action='append',
        help=(
            'If provided, at least one of these regular expressions must '
            'match for a string to be included in the database.'
        ),
    )
    subparser.add_argument(
        '-e',
        '--exclude',
        type=re.compile,
        default=[],
        action='append',
        help=(
            'If provided, none of these regular expressions may match for a '
            'string to be included in the database.'
        ),
    )

    unescaped_slash = re.compile(r'(?<!\\)/')

    def replacement(value: str) -> Tuple[Pattern, 'str']:
        try:
            find, sub = unescaped_slash.split(value, 1)
        except ValueError as _err:
            raise argparse.ArgumentTypeError(
                'replacements must be specified as "search_regex/replacement"'
            )

        try:
            return re.compile(find.replace(r'\/', '/')), sub
        except re.error as err:
            raise argparse.ArgumentTypeError(
                f'"{value}" is not a valid regular expression: {err}'
            )

    subparser.add_argument(
        '--replace',
        type=replacement,
        default=[],
        action='append',
        help=(
            'If provided, replaces text that matches a regular expression. '
            'This can be used to replace sensitive terms in a token '
            'database that will be distributed publicly. The expression and '
            'replacement are specified as "search_regex/replacement". '
            'Plain slash characters in the regex must be escaped with a '
            r'backslash (\/). The replacement text may include '
            'backreferences for captured groups in the regex.'
        ),
    )

    # The 'add' command adds strings to a database from a set of ELFs.
    subparser = subparsers.add_parser(
        'add',
        parents=[option_db, option_tokens],
        help=(
            'Adds new strings to a database with tokenized strings from a set '
            'of ELF files or other token databases. Missing entries are NOT '
            'marked as removed.'
        ),
    )
    subparser.set_defaults(handler=_handle_add)
    subparser.add_argument(
        '--discard-temporary',
        dest='commit',
        help=(
            'Deletes temporary tokens in memory and on disk when a CSV exists '
            'within a commit. Afterwards, new strings are added to the '
            'database from a set of ELF files or other token databases. '
            'Missing entries are NOT marked as removed.'
        ),
    )

    # The 'mark_removed' command marks removed entries to match a set of ELFs.
    subparser = subparsers.add_parser(
        'mark_removed',
        parents=[option_db, option_tokens],
        help=(
            'Updates a database with tokenized strings from a set of strings. '
            'Strings not present in the set remain in the database but are '
            'marked as removed. New strings are NOT added.'
        ),
    )
    subparser.set_defaults(handler=_handle_mark_removed)
    subparser.add_argument(
        '--date',
        type=year_month_day,
        help=(
            'The removal date to use for all strings. '
            'May be YYYY-MM-DD or "today". (default: today)'
        ),
    )

    # The 'purge' command removes old entries.
    subparser = subparsers.add_parser(
        'purge',
        parents=[option_db],
        help='Purges removed strings from a database.',
    )
    subparser.set_defaults(handler=_handle_purge)
    subparser.add_argument(
        '-b',
        '--before',
        type=year_month_day,
        help=(
            'Delete all entries removed on or before this date. '
            'May be YYYY-MM-DD or "today".'
        ),
    )

    # The 'report' command prints a report about a database.
    subparser = subparsers.add_parser(
        'report', help='Prints a report about a database.'
    )
    subparser.set_defaults(handler=_handle_report)
    subparser.add_argument(
        'token_database_or_elf',
        nargs='+',
        action=ExpandGlobs,
        help=(
            'The ELF files or token databases about which to generate '
            'reports.'
        ),
    )
    subparser.add_argument(
        '-o',
        '--output',
        type=argparse.FileType('w'),
        default=sys.stdout,
        help='The file to which to write the output; use - for stdout.',
    )

    args = parser.parse_args()

    handler = args.handler
    del args.handler

    return handler, args


def _init_logging(level: int) -> None:
    _LOG.setLevel(logging.DEBUG)
    log_to_stderr = logging.StreamHandler()
    log_to_stderr.setLevel(level)
    log_to_stderr.setFormatter(
        logging.Formatter(
            fmt='%(asctime)s.%(msecs)03d-%(levelname)s: %(message)s',
            datefmt='%H:%M:%S',
        )
    )

    _LOG.addHandler(log_to_stderr)


def _main(handler: Callable, args: argparse.Namespace) -> int:
    _init_logging(logging.INFO)
    handler(**vars(args))
    return 0


if __name__ == '__main__':
    sys.exit(_main(*_parse_args()))



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2020 The Pigweed Authors

Licensed under the Apache License, Version 2.0 (the "License"); you may not
use this file except in compliance with the License. You may obtain a copy of
the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class ArmV7mCpuState(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PC_FIELD_NUMBER: builtins.int
    LR_FIELD_NUMBER: builtins.int
    PSR_FIELD_NUMBER: builtins.int
    MSP_FIELD_NUMBER: builtins.int
    PSP_FIELD_NUMBER: builtins.int
    EXC_RETURN_FIELD_NUMBER: builtins.int
    CFSR_FIELD_NUMBER: builtins.int
    MSPLIM_FIELD_NUMBER: builtins.int
    PSPLIM_FIELD_NUMBER: builtins.int
    MMFAR_FIELD_NUMBER: builtins.int
    BFAR_FIELD_NUMBER: builtins.int
    ICSR_FIELD_NUMBER: builtins.int
    HFSR_FIELD_NUMBER: builtins.int
    SHCSR_FIELD_NUMBER: builtins.int
    CONTROL_FIELD_NUMBER: builtins.int
    R0_FIELD_NUMBER: builtins.int
    R1_FIELD_NUMBER: builtins.int
    R2_FIELD_NUMBER: builtins.int
    R3_FIELD_NUMBER: builtins.int
    R4_FIELD_NUMBER: builtins.int
    R5_FIELD_NUMBER: builtins.int
    R6_FIELD_NUMBER: builtins.int
    R7_FIELD_NUMBER: builtins.int
    R8_FIELD_NUMBER: builtins.int
    R9_FIELD_NUMBER: builtins.int
    R10_FIELD_NUMBER: builtins.int
    R11_FIELD_NUMBER: builtins.int
    R12_FIELD_NUMBER: builtins.int
    pc: builtins.int
    lr: builtins.int
    psr: builtins.int
    msp: builtins.int
    psp: builtins.int
    exc_return: builtins.int
    cfsr: builtins.int
    msplim: builtins.int
    psplim: builtins.int
    mmfar: builtins.int
    bfar: builtins.int
    icsr: builtins.int
    hfsr: builtins.int
    shcsr: builtins.int
    control: builtins.int
    r0: builtins.int
    """General purpose registers."""
    r1: builtins.int
    r2: builtins.int
    r3: builtins.int
    r4: builtins.int
    r5: builtins.int
    r6: builtins.int
    r7: builtins.int
    r8: builtins.int
    r9: builtins.int
    r10: builtins.int
    r11: builtins.int
    r12: builtins.int
    def __init__(
        self,
        *,
        pc: builtins.int | None = ...,
        lr: builtins.int | None = ...,
        psr: builtins.int | None = ...,
        msp: builtins.int | None = ...,
        psp: builtins.int | None = ...,
        exc_return: builtins.int | None = ...,
        cfsr: builtins.int | None = ...,
        msplim: builtins.int | None = ...,
        psplim: builtins.int | None = ...,
        mmfar: builtins.int | None = ...,
        bfar: builtins.int | None = ...,
        icsr: builtins.int | None = ...,
        hfsr: builtins.int | None = ...,
        shcsr: builtins.int | None = ...,
        control: builtins.int | None = ...,
        r0: builtins.int | None = ...,
        r1: builtins.int | None = ...,
        r2: builtins.int | None = ...,
        r3: builtins.int | None = ...,
        r4: builtins.int | None = ...,
        r5: builtins.int | None = ...,
        r6: builtins.int | None = ...,
        r7: builtins.int | None = ...,
        r8: builtins.int | None = ...,
        r9: builtins.int | None = ...,
        r10: builtins.int | None = ...,
        r11: builtins.int | None = ...,
        r12: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["bfar", b"bfar", "cfsr", b"cfsr", "control", b"control", "exc_return", b"exc_return", "hfsr", b"hfsr", "icsr", b"icsr", "lr", b"lr", "mmfar", b"mmfar", "msp", b"msp", "msplim", b"msplim", "pc", b"pc", "psp", b"psp", "psplim", b"psplim", "psr", b"psr", "r0", b"r0", "r1", b"r1", "r10", b"r10", "r11", b"r11", "r12", b"r12", "r2", b"r2", "r3", b"r3", "r4", b"r4", "r5", b"r5", "r6", b"r6", "r7", b"r7", "r8", b"r8", "r9", b"r9", "shcsr", b"shcsr"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["bfar", b"bfar", "cfsr", b"cfsr", "control", b"control", "exc_return", b"exc_return", "hfsr", b"hfsr", "icsr", b"icsr", "lr", b"lr", "mmfar", b"mmfar", "msp", b"msp", "msplim", b"msplim", "pc", b"pc", "psp", b"psp", "psplim", b"psplim", "psr", b"psr", "r0", b"r0", "r1", b"r1", "r10", b"r10", "r11", b"r11", "r12", b"r12", "r2", b"r2", "r3", b"r3", "r4", b"r4", "r5", b"r5", "r6", b"r6", "r7", b"r7", "r8", b"r8", "r9", b"r9", "shcsr", b"shcsr"]) -> None: ...

global___ArmV7mCpuState = ArmV7mCpuState

class SnapshotCpuStateOverlay(google.protobuf.message.Message):
    """This message overlays the pw.snapshot.Snapshot proto. It's valid to encode
    this message to the same sink that a Snapshot proto is being written to.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ARMV7M_CPU_STATE_FIELD_NUMBER: builtins.int
    @property
    def armv7m_cpu_state(self) -> global___ArmV7mCpuState:
        """TODO(amontanez): Rename this to be Cortex-M arch generic, e.g.:
          optional CpuState cpu_state = 20;
        """
    def __init__(
        self,
        *,
        armv7m_cpu_state: global___ArmV7mCpuState | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["armv7m_cpu_state", b"armv7m_cpu_state"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["armv7m_cpu_state", b"armv7m_cpu_state"]) -> None: ...

global___SnapshotCpuStateOverlay = SnapshotCpuStateOverlay



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""A simple tool to decode a CFSR register from the command line

Example usage:

  $ python -m pw_cpu_exception_cortex_m.cfsr_decoder 0x00010100

  20210412 15:09:01 INF Exception caused by a usage fault, bus fault.

  Active Crash Fault Status Register (CFSR) fields:
  IBUSERR     Bus fault on instruction fetch.
  UNDEFINSTR  Encountered invalid instruction.

  All registers:
  cfsr       0x00010100
"""

import argparse
import logging
import sys
import pw_cli.log

from pw_cpu_exception_cortex_m_protos import cpu_state_pb2
from pw_cpu_exception_cortex_m import exception_analyzer

_LOG = logging.getLogger('decode_cfsr')


def _parse_args() -> argparse.Namespace:
    """Parses arguments for this script, splitting out the command to run."""
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        'cfsr', type=lambda val: int(val, 0), help='The Cortex-M CFSR to decode'
    )
    return parser.parse_args()


def dump_cfsr(cfsr: int) -> int:
    cpu_state_proto = cpu_state_pb2.ArmV7mCpuState()
    cpu_state_proto.cfsr = cfsr
    cpu_state_info = exception_analyzer.CortexMExceptionAnalyzer(
        cpu_state_proto
    )
    _LOG.info(cpu_state_info)
    return 0


if __name__ == '__main__':
    pw_cli.log.install(level=logging.INFO)
    sys.exit(dump_cfsr(**vars(_parse_args())))



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Provides functionality for encoding tokenized messages."""

import argparse
import base64
import struct
import sys
from typing import Sequence, Union

from pw_tokenizer import tokens

_INT32_MAX = 2**31 - 1
_UINT32_MAX = 2**32 - 1
BASE64_PREFIX = '$'


def _zig_zag_encode(value: int) -> int:
    """Encodes signed integers to give a compact varint encoding."""
    return value << 1 if value >= 0 else (value << 1) ^ (~0)


def _little_endian_base128_encode(integer: int) -> bytearray:
    data = bytearray()

    while True:
        # Grab 7 bits; the eighth bit is set to 1 to indicate more data coming.
        data.append((integer & 0x7F) | 0x80)
        integer >>= 7

        if not integer:
            break

    data[-1] &= 0x7F  # clear the top bit of the last byte
    return data


def _encode_int32(arg: int) -> bytearray:
    # Convert large unsigned numbers into their corresponding signed values.
    if arg > _INT32_MAX:
        arg -= 2**32

    return _little_endian_base128_encode(_zig_zag_encode(arg))


def _encode_string(arg: bytes) -> bytes:
    size_byte = len(arg) if len(arg) < 128 else 0xFF
    return struct.pack('B', size_byte) + arg[:127]


def encode_args(*args: Union[int, float, bytes, str]) -> bytes:
    """Encodes a list of arguments to their on-wire representation."""

    data = bytearray(b'')
    for arg in args:
        if isinstance(arg, int):
            if arg.bit_length() > 32:
                raise ValueError(
                    f'Cannot encode {arg}: only 32-bit integers may be encoded'
                )
            data += _encode_int32(arg)
        elif isinstance(arg, float):
            data += struct.pack('<f', arg)
        elif isinstance(arg, str):
            data += _encode_string(arg.encode())
        elif isinstance(arg, bytes):
            data += _encode_string(arg)
        else:
            raise ValueError(
                f'{arg} has type {type(arg)}, which is not supported'
            )
    return bytes(data)


def encode_token_and_args(
    token: int, *args: Union[int, float, bytes, str]
) -> bytes:
    """Encodes a tokenized message given its token and arguments.

    This function assumes that the token represents a format string with
    conversion specifiers that correspond with the provided argument types.
    Currently, only 32-bit integers are supported.
    """

    if token < 0 or token > _UINT32_MAX:
        raise ValueError(
            f'The token ({token}) must be an unsigned 32-bit integer'
        )

    return struct.pack('<I', token) + encode_args(*args)


def prefixed_base64(data: bytes, prefix: str = '$') -> str:
    """Encodes a tokenized message as prefixed Base64."""
    return prefix + base64.b64encode(data).decode()


def _parse_user_input(string: str):
    """Evaluates a string as Python code or returns it as a literal string."""
    try:
        value = eval(string, dict(__builtins__={}))  # pylint: disable=eval-used
    except (NameError, SyntaxError):
        return string

    return value if isinstance(value, (int, float)) else string


def _main(format_string_list: Sequence[str], raw_args: Sequence[str]) -> int:
    (format_string,) = format_string_list
    token = tokens.pw_tokenizer_65599_hash(format_string)
    args = tuple(_parse_user_input(a) for a in raw_args)

    data = encode_token_and_args(token, *args)
    token = int.from_bytes(data[:4], 'little')
    binary = ' '.join(f'{b:02x}' for b in data)

    print(f'      Raw input: {format_string!r} % {args!r}')
    print(f'Formatted input: {format_string % args}')
    print(f'          Token: 0x{token:08x}')
    print(f'        Encoded: {data!r} ({binary}) [{len(data)} bytes]')
    print(f'Prefixed Base64: {prefixed_base64(data)}')

    return 0


def _parse_args() -> dict:
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        'format_string_list',
        metavar='FORMAT_STRING',
        nargs=1,
        help='Format string with optional %%-style arguments.',
    )
    parser.add_argument(
        'raw_args',
        metavar='ARG',
        nargs='*',
        help=(
            'Arguments for the format string, if any. Arguments are parsed '
            'as Python expressions, with no builtins (e.g. 9 is the number '
            '9 and \'"9"\' is the string "9"). Arguments that are not valid '
            'Python are treated as string literals.'
        ),
    )
    return vars(parser.parse_args())


if __name__ == '__main__':
    sys.exit(_main(**_parse_args()))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_chrono_protos/chrono.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_chrono_protos/chrono.proto',
  package='pw.chrono',
  syntax='proto3',
  serialized_options=b'B\006Chrono',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1dpw_chrono_protos/chrono.proto\x12\tpw.chrono\"q\n\tEpochType\"d\n\x04\x45num\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x13\n\x0fTIME_SINCE_BOOT\x10\x01\x12\x12\n\x0eUTC_WALL_CLOCK\x10\x02\x12\x12\n\x0eGPS_WALL_CLOCK\x10\x03\x12\x12\n\x0eTAI_WALL_CLOCK\x10\x04\"\xa4\x01\n\x0f\x43lockParameters\x12%\n\x1dtick_period_seconds_numerator\x18\x01 \x01(\x05\x12\'\n\x1ftick_period_seconds_denominator\x18\x02 \x01(\x05\x12\x32\n\nepoch_type\x18\x03 \x01(\x0e\x32\x19.pw.chrono.EpochType.EnumH\x00\x88\x01\x01\x42\r\n\x0b_epoch_type\"T\n\tTimePoint\x12\x11\n\ttimestamp\x18\x01 \x01(\x03\x12\x34\n\x10\x63lock_parameters\x18\x02 \x01(\x0b\x32\x1a.pw.chrono.ClockParameters\">\n\x12SnapshotTimestamps\x12(\n\ntimestamps\x18\x16 \x03(\x0b\x32\x14.pw.chrono.TimePointB\x08\x42\x06\x43hronob\x06proto3'
)



_EPOCHTYPE_ENUM = _descriptor.EnumDescriptor(
  name='Enum',
  full_name='pw.chrono.EpochType.Enum',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TIME_SINCE_BOOT', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UTC_WALL_CLOCK', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GPS_WALL_CLOCK', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TAI_WALL_CLOCK', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=57,
  serialized_end=157,
)
_sym_db.RegisterEnumDescriptor(_EPOCHTYPE_ENUM)


_EPOCHTYPE = _descriptor.Descriptor(
  name='EpochType',
  full_name='pw.chrono.EpochType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _EPOCHTYPE_ENUM,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=44,
  serialized_end=157,
)


_CLOCKPARAMETERS = _descriptor.Descriptor(
  name='ClockParameters',
  full_name='pw.chrono.ClockParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='tick_period_seconds_numerator', full_name='pw.chrono.ClockParameters.tick_period_seconds_numerator', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tick_period_seconds_denominator', full_name='pw.chrono.ClockParameters.tick_period_seconds_denominator', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='epoch_type', full_name='pw.chrono.ClockParameters.epoch_type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_epoch_type', full_name='pw.chrono.ClockParameters._epoch_type',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=160,
  serialized_end=324,
)


_TIMEPOINT = _descriptor.Descriptor(
  name='TimePoint',
  full_name='pw.chrono.TimePoint',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='pw.chrono.TimePoint.timestamp', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='clock_parameters', full_name='pw.chrono.TimePoint.clock_parameters', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=326,
  serialized_end=410,
)


_SNAPSHOTTIMESTAMPS = _descriptor.Descriptor(
  name='SnapshotTimestamps',
  full_name='pw.chrono.SnapshotTimestamps',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamps', full_name='pw.chrono.SnapshotTimestamps.timestamps', index=0,
      number=22, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=412,
  serialized_end=474,
)

_EPOCHTYPE_ENUM.containing_type = _EPOCHTYPE
_CLOCKPARAMETERS.fields_by_name['epoch_type'].enum_type = _EPOCHTYPE_ENUM
_CLOCKPARAMETERS.oneofs_by_name['_epoch_type'].fields.append(
  _CLOCKPARAMETERS.fields_by_name['epoch_type'])
_CLOCKPARAMETERS.fields_by_name['epoch_type'].containing_oneof = _CLOCKPARAMETERS.oneofs_by_name['_epoch_type']
_TIMEPOINT.fields_by_name['clock_parameters'].message_type = _CLOCKPARAMETERS
_SNAPSHOTTIMESTAMPS.fields_by_name['timestamps'].message_type = _TIMEPOINT
DESCRIPTOR.message_types_by_name['EpochType'] = _EPOCHTYPE
DESCRIPTOR.message_types_by_name['ClockParameters'] = _CLOCKPARAMETERS
DESCRIPTOR.message_types_by_name['TimePoint'] = _TIMEPOINT
DESCRIPTOR.message_types_by_name['SnapshotTimestamps'] = _SNAPSHOTTIMESTAMPS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

EpochType = _reflection.GeneratedProtocolMessageType('EpochType', (_message.Message,), {
  'DESCRIPTOR' : _EPOCHTYPE,
  '__module__' : 'pw_chrono_protos.chrono_pb2'
  # @@protoc_insertion_point(class_scope:pw.chrono.EpochType)
  })
_sym_db.RegisterMessage(EpochType)

ClockParameters = _reflection.GeneratedProtocolMessageType('ClockParameters', (_message.Message,), {
  'DESCRIPTOR' : _CLOCKPARAMETERS,
  '__module__' : 'pw_chrono_protos.chrono_pb2'
  # @@protoc_insertion_point(class_scope:pw.chrono.ClockParameters)
  })
_sym_db.RegisterMessage(ClockParameters)

TimePoint = _reflection.GeneratedProtocolMessageType('TimePoint', (_message.Message,), {
  'DESCRIPTOR' : _TIMEPOINT,
  '__module__' : 'pw_chrono_protos.chrono_pb2'
  # @@protoc_insertion_point(class_scope:pw.chrono.TimePoint)
  })
_sym_db.RegisterMessage(TimePoint)

SnapshotTimestamps = _reflection.GeneratedProtocolMessageType('SnapshotTimestamps', (_message.Message,), {
  'DESCRIPTOR' : _SNAPSHOTTIMESTAMPS,
  '__module__' : 'pw_chrono_protos.chrono_pb2'
  # @@protoc_insertion_point(class_scope:pw.chrono.SnapshotTimestamps)
  })
_sym_db.RegisterMessage(SnapshotTimestamps)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Serializes an Environment into a batch file."""

# Disable super() warnings since this file must be Python 2 compatible.
# pylint: disable=super-with-arguments

# goto label written to the end of Windows batch files for exiting a script.
_SCRIPT_END_LABEL = '_pw_end'


class BatchVisitor(object):  # pylint: disable=useless-object-inheritance
    """Serializes an Environment into a batch file."""

    def __init__(self, *args, **kwargs):
        pathsep = kwargs.pop('pathsep', ':')
        super(BatchVisitor, self).__init__(*args, **kwargs)
        self._replacements = ()
        self._outs = None
        self._pathsep = pathsep

    def serialize(self, env, outs):
        try:
            self._replacements = tuple(
                (key, env.get(key) if value is None else value)
                for key, value in env.replacements
            )
            self._outs = outs
            self._outs.write('@echo off\n')

            env.accept(self)

            outs.write(':{}\n'.format(_SCRIPT_END_LABEL))

        finally:
            self._replacements = ()
            self._outs = None

    def _apply_replacements(self, action):
        value = action.value
        for var, replacement in self._replacements:
            if var != action.name:
                value = value.replace(replacement, '%{}%'.format(var))
        return value

    def visit_set(self, set):  # pylint: disable=redefined-builtin
        value = self._apply_replacements(set)
        self._outs.write(
            'set {name}={value}\n'.format(name=set.name, value=value)
        )

    def visit_clear(self, clear):
        self._outs.write('set {name}=\n'.format(name=clear.name))

    def visit_remove(self, remove):
        pass  # Not supported on Windows.

    def _join(self, *args):
        if len(args) == 1 and isinstance(args[0], (list, tuple)):
            args = args[0]
        return self._pathsep.join(args)

    def visit_prepend(self, prepend):
        value = self._apply_replacements(prepend)
        value = self._join(value, '%{}%'.format(prepend.name))
        self._outs.write(
            'set {name}={value}\n'.format(name=prepend.name, value=value)
        )

    def visit_append(self, append):
        value = self._apply_replacements(append)
        value = self._join('%{}%'.format(append.name), value)
        self._outs.write(
            'set {name}={value}\n'.format(name=append.name, value=value)
        )

    def visit_echo(self, echo):
        if echo.newline:
            if not echo.value:
                self._outs.write('echo.\n')
            else:
                self._outs.write('echo {}\n'.format(echo.value))
        else:
            self._outs.write('<nul set /p="{}"\n'.format(echo.value))

    def visit_comment(self, comment):
        for line in comment.value.splitlines():
            self._outs.write(':: {}\n'.format(line))

    def visit_command(self, command):
        # TODO(mohrr) use shlex.quote here?
        self._outs.write('{}\n'.format(' '.join(command.command)))
        if not command.exit_on_error:
            return

        # Assume failing command produced relevant output.
        self._outs.write(
            'if %ERRORLEVEL% neq 0 goto {}\n'.format(_SCRIPT_END_LABEL)
        )

    def visit_doctor(self, doctor):
        self._outs.write('if "%PW_ACTIVATE_SKIP_CHECKS%"=="" (\n')
        self.visit_command(doctor)
        self._outs.write(') else (\n')
        self._outs.write(
            'echo Skipping environment check because '
            'PW_ACTIVATE_SKIP_CHECKS is set\n'
        )
        self._outs.write(')\n')

    def visit_blank_line(self, blank_line):
        del blank_line
        self._outs.write('\n')

    def visit_function(self, function):
        pass  # Not supported on Windows.

    def visit_hash(self, hash):  # pylint: disable=redefined-builtin
        pass  # Not relevant on Windows.



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Yaml config file loader mixin."""

import os
import logging
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import yaml

_LOG = logging.getLogger(__package__)


class MissingConfigTitle(Exception):
    """Exception for when an existing YAML file is missing config_title."""


class YamlConfigLoaderMixin:
    """Yaml Config file loader mixin.

    Use this mixin to load yaml file settings and save them into
    ``self._config``. For example:

    ::

       class ConsolePrefs(YamlConfigLoaderMixin):
           def __init__(self) -> None:
               self.config_init(
                   config_section_title='pw_console',
                   project_file=Path('project_file.yaml'),
                   project_user_file=Path('project_user_file.yaml'),
                   user_file=Path('~/user_file.yaml'),
                   default_config={},
                   environment_var='PW_CONSOLE_CONFIG_FILE',
               )

    """

    def config_init(
        self,
        config_section_title: str,
        project_file: Optional[Union[Path, bool]] = None,
        project_user_file: Optional[Union[Path, bool]] = None,
        user_file: Optional[Union[Path, bool]] = None,
        default_config: Optional[Dict[Any, Any]] = None,
        environment_var: Optional[str] = None,
    ) -> None:
        """Call this to load YAML config files in order of precedence.

        The following files are loaded in this order:
        1. project_file
        2. project_user_file
        3. user_file

        Lastly, if a valid file path is specified at
        ``os.environ[environment_var]`` then load that file overriding all
        config options.

        Args:
            config_section_title: String name of this config section. For
                example: ``pw_console`` or ``pw_watch``. In the YAML file this
                is represented by a ``config_title`` key.

                ::

                   ---
                   config_title: pw_console

            project_file: Project level config file. This is intended to be a
                file living somewhere under a project folder and is checked into
                the repo. It serves as a base config all developers can inherit
                from.
            project_user_file: User's personal config file for a specific
                project. This can be a file that lives in a project folder that
                is git-ignored and not checked into the repo.
            user_file: A global user based config file. This is typically a file
                in the users home directory and settings here apply to all
                projects.
            default_config: A Python dict representing the base default
                config. This dict will be applied as a starting point before
                loading any yaml files.
            environment_var: The name of an environment variable to check for a
                config file. If a config file exists there it will be loaded on
                top of the default_config ignoring project and user files.
        """

        self._config_section_title: str = config_section_title
        self.default_config = default_config if default_config else {}
        self.reset_config()

        if project_file and isinstance(project_file, Path):
            self.project_file = Path(
                os.path.expandvars(str(project_file.expanduser()))
            )
            self.load_config_file(self.project_file)

        if project_user_file and isinstance(project_user_file, Path):
            self.project_user_file = Path(
                os.path.expandvars(str(project_user_file.expanduser()))
            )
            self.load_config_file(self.project_user_file)

        if user_file and isinstance(user_file, Path):
            self.user_file = Path(
                os.path.expandvars(str(user_file.expanduser()))
            )
            self.load_config_file(self.user_file)

        # Check for a config file specified by an environment variable.
        if environment_var is None:
            return
        environment_config = os.environ.get(environment_var, None)
        if environment_config:
            env_file_path = Path(environment_config)
            if not env_file_path.is_file():
                raise FileNotFoundError(
                    f'Cannot load config file: {env_file_path}'
                )
            self.reset_config()
            self.load_config_file(env_file_path)

    def _update_config(self, cfg: Dict[Any, Any]) -> None:
        if cfg is None:
            cfg = {}
        self._config.update(cfg)

    def reset_config(self) -> None:
        self._config: Dict[Any, Any] = {}
        self._update_config(self.default_config)

    def _load_config_from_string(  # pylint: disable=no-self-use
        self, file_contents: str
    ) -> List[Dict[Any, Any]]:
        return list(yaml.safe_load_all(file_contents))

    def load_config_file(self, file_path: Path) -> None:
        if not file_path.is_file():
            return

        cfgs = self._load_config_from_string(file_path.read_text())

        for cfg in cfgs:
            if self._config_section_title in cfg:
                self._update_config(cfg[self._config_section_title])

            elif cfg.get('config_title', False) == self._config_section_title:
                self._update_config(cfg)
            else:
                raise MissingConfigTitle(
                    '\n\nThe config file "{}" is missing the expected '
                    '"config_title: {}" setting.'.format(
                        str(file_path), self._config_section_title
                    )
                )



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Decodes arguments and formats tokenized messages.

The decode(format_string, encoded_arguments) function provides a simple way to
format a string with encoded arguments. The FormatString class may also be used.

Missing, truncated, or otherwise corrupted arguments are handled and displayed
in the resulting string with an error message.
"""

from datetime import datetime
import math
import re
import struct
from typing import (
    Iterable,
    List,
    NamedTuple,
    Match,
    Optional,
    Sequence,
    Tuple,
)


def zigzag_decode(value: int) -> int:
    """ZigZag decode function from protobuf's wire_format module."""
    if not value & 0x1:
        return value >> 1
    return (value >> 1) ^ (~0)


class FormatSpec:
    """Represents a format specifier parsed from a printf-style string.

    This implementation is designed to align with the C99 specification,
    section 7.19.6
    (https://www.dii.uchile.cl/~daespino/files/Iso_C_1999_definition.pdf).
    Notably, this specification is slightly different than what is implemented
    in most compilers due to each compiler choosing to interpret undefined
    behavior in slightly different ways. Treat the following description as the
    source of truth.

    This implementation supports:
    - Overall Format: `%[flags][width][.precision][length][specifier]`
    - Flags (Zero or More)
      - `-`: Left-justify within the given field width; Right justification is
             the default (see Width modifier).
      - `+`: Forces to preceed the result with a plus or minus sign (`+` or `-`)
             even for positive numbers. By default, only negative numbers are
             preceded with a `-` sign.
      - ` ` (space): If no sign is going to be written, a blank space is
                     inserted before the value.
      - `#`: Specifies an alternative print syntax should be used.
        - Used with `o`, `x` or `X` specifiers the value is preceeded with `0`,
          `0x`, or `0X`, respectively, for values different than zero.
        - Used with `a`, `A`, `e`, `E`, `f`, `F`, `g`, or `G` it forces the
          written output to contain a decimal point even if no more digits
          follow. By default, if no digits follow, no decimal point is written.
      - `0`: Left-pads the number with zeroes (`0`) instead of spaces when
             padding is specified (see width sub-specifier).
    - Width (Optional)
      - ``(number)``: Minimum number of characters to be printed. If the value
                      to be printed is shorter than this number, the result is
                      padded with blank spaces or `0` if the `0` flag is
                      present. The value is not truncated even if the result is
                      larger. If the value is negative and the `0` flag is
                      present, the `0`s are padded after the `-` symbol.
      - `*`: The width is not specified in the format string, but as an
             additional integer value argument preceding the argument that has
             to be formatted.
    - Precision (Optional)
      - `.(number)`
        - For `d`, `i`, `o`, `u`, `x`, `X`, specifies the minimum number of
          digits to be written. If the value to be written is shorter than this
          number, the result is padded with leading zeros. The value is not
          truncated even if the result is longer.
          - A precision of `0` means that no character is written for the value
            `0`.
        - For `a`, `A`, `e`, `E`, `f`, and `F`, specifies the number of digits
          to be printed after the decimal point. By default, this is `6`.
        - For `g` and `G`, specifies the maximum number of significant digits to
          be printed.
        - For `s`, specifies the maximum number of characters to be printed. By
          default all characters are printed until the ending null character is
          encountered.
        - If the period is specified without an explicit value for precision,
          `0` is assumed.
      - `.*`: The precision is not specified in the format string, but as an
              additional integer value argument preceding the argument that has
              to be formatted.
    - Length (Optional)
      - `hh`: Usable with `d`, `i`, `o`, `u`, `x`, or `X` specifiers to convey
              the argument will be a `signed char` or `unsigned char`. However,
              this is largely ignored in the implementation due to it not being
              necessary for Python or argument decoding (since the argument is
              always encoded at least as a 32-bit integer).
      - `h`: Usable with `d`, `i`, `o`, `u`, `x`, or `X` specifiers to convey
             the argument will be a `signed short int` or `unsigned short int`.
             However, this is largely ignored in the implementation due to it
             not being necessary for Python or argument decoding (since the
             argument is always encoded at least as a 32-bit integer).
      - `l`: Usable with `d`, `i`, `o`, `u`, `x`, or `X` specifiers to convey
             the argument will be a `signed long int` or `unsigned long int`.
             Also is usable with `c` and `s` to specify that the arguments will
             be encoded with `wchar_t` values (which isn't different from normal
             `char` values). However, this is largely ignored in the
             implementation due to it not being necessary for Python or argument
             decoding (since the argument is always encoded at least as a 32-bit
             integer).
      - `ll`: Usable with `d`, `i`, `o`, `u`, `x`, or `X` specifiers to convey
              the argument will be a `signed long long int` or
              `unsigned long long int`. This is required to properly decode the
              argument as a 64-bit integer.
      - `L`: Usable with `a`, `A`, `e`, `E`, `f`, `F`, `g`, or `G` conversion
             specifiers applies to a long double argument. However, this is
             ignored in the implementation due to floating point value encoded
             that is unaffected by bit width.
      - `j`: Usable with `d`, `i`, `o`, `u`, `x`, or `X` specifiers to convey
             the argument will be a `intmax_t` or `uintmax_t`.
      - `z`: Usable with `d`, `i`, `o`, `u`, `x`, or `X` specifiers to convey
             the argument will be a `size_t`. This will force the argument to be
             decoded as an unsigned integer.
      - `t`: Usable with `d`, `i`, `o`, `u`, `x`, or `X` specifiers to convey
             the argument will be a `ptrdiff_t`.
      - If a length modifier is provided for an incorrect specifier, it is
        ignored.
    - Specifier (Required)
      - `d` / `i`: Used for signed decimal integers.
      - `u`: Used for unsigned decimal integers.
      - `o`: Used for unsigned decimal integers and specifies formatting should
             be as an octal number.
      - `x`: Used for unsigned decimal integers and specifies formatting should
             be as a hexadecimal number using all lowercase letters.
      - `X`: Used for unsigned decimal integers and specifies formatting should
             be as a hexadecimal number using all uppercase letters.
      - `f`: Used for floating-point values and specifies to use lowercase,
             decimal floating point formatting.
        - Default precision is `6` decimal places unless explicitly specified.
      - `F`: Used for floating-point values and specifies to use uppercase,
             decimal floating point formatting.
        - Default precision is `6` decimal places unless explicitly specified.
      - `e`: Used for floating-point values and specifies to use lowercase,
             exponential (scientific) formatting.
        - Default precision is `6` decimal places unless explicitly specified.
      - `E`: Used for floating-point values and specifies to use uppercase,
             exponential (scientific) formatting.
        - Default precision is `6` decimal places unless explicitly specified.
      - `g`: Used for floating-point values and specified to use `f` or `e`
             formatting depending on which would be the shortest representation.
        - Precision specifies the number of significant digits, not just digits
          after the decimal place.
        - If the precision is specified as `0`, it is interpreted to mean `1`.
        - `e` formatting is used if the the exponent would be less than `-4` or
          is greater than or equal to the precision.
        - Trailing zeros are removed unless the `#` flag is set.
        - A decimal point only appears if it is followed by a digit.
        - `NaN` or infinities always follow `f` formatting.
      - `G`: Used for floating-point values and specified to use `f` or `e`
             formatting depending on which would be the shortest representation.
        - Precision specifies the number of significant digits, not just digits
          after the decimal place.
        - If the precision is specified as `0`, it is interpreted to mean `1`.
        - `E` formatting is used if the the exponent would be less than `-4` or
          is greater than or equal to the precision.
        - Trailing zeros are removed unless the `#` flag is set.
        - A decimal point only appears if it is followed by a digit.
        - `NaN` or infinities always follow `F` formatting.
      - `c`: Used for formatting a `char` value.
      - `s`: Used for formatting a string of `char` values.
        - If width is specified, the null terminator character is included as a
          character for width count.
        - If precision is specified, no more `char`s than that value will be
          written from the string (padding is used to fill additional width).
      - `p`: Used for formatting a pointer address.
      - `%`: Prints a single `%`. Only valid as `%%` (supports no flags, width,
             precision, or length modifiers).

    Underspecified details:
    - If both `+` and ` ` flags appear, the ` ` is ignored.
    - The `+` and ` ` flags will error if used with `c` or `s`.
    - The `#` flag will error if used with `d`, `i`, `u`, `c`, `s`, or `p`.
    - The `0` flag will error if used with `c`, `s`, or `p`.
    - Both `+` and ` ` can work with the unsigned integer specifiers `u`, `o`,
      `x`, and `X`.
    - If a length modifier is provided for an incorrect specifier, it is
      ignored.
    - The `z` length modifier will decode arugments as signed as long as `d` or
      `i` is used.
    - `p` is implementation defined. For this implementation, it will print
      with a `0x` prefix and then the pointer value was printed using `%08X`.
      `p` supports the `+`, `-`, and ` ` flags, but not the `#` or `0` flags.
      None of the length modifiers are usable with `p`. This implementation will
      try to adhere to user-specified width (assuming the width provided is
      larger than the guaranteed minimum of 10). Specifying precision for `p` is
      considered an error.
    - Only `%%` is allowed with no other modifiers. Things like `%+%` will fail
      to decode. Some C stdlib implementations support any modifiers being
      present between `%`, but ignore any for the output.
    - If a width is specified with the `0` flag for a negative value, the padded
      `0`s will appear after the `-` symbol.
    - A precision of `0` for `d`, `i`, `u`, `o`, `x`, or `X` means that no
      character is written for the value `0`.
    - Precision cannot be specified for `c`.
    - Using `*` or fixed precision with the `s` specifier still requires the
      string argument to be null-terminated. This is due to argument encoding
      happening on the C/C++-side while the precision value is not read or
      otherwise used until decoding happens in this Python code.

    Non-conformant details:
    - `n` specifier: We do not support the `n` specifier since it is impossible
                     for us to retroactively tell the original program how many
                     characters have been printed since this decoding happens a
                     great deal of time after the device sent it, usually on a
                     separate processing device entirely.
    """

    # Regular expression for finding format specifiers.
    FORMAT_SPEC = re.compile(
        r'%(?P<flags>[+\- #0]+)?'
        r'(?P<width>\d+|\*)?'
        r'(?P<precision>\.(?:\d*|\*))?'
        r'(?P<length>hh|h|ll|l|j|z|t|L)?'
        r'(?P<type>[csdioxXufFeEaAgGnp%])'
    )

    # Conversions to make format strings Python compatible.
    _REMAP_TYPE = {'a': 'f', 'A': 'F', 'p': 'X'}

    # Conversion specifiers by type; n is not supported.
    SIGNED_INT = frozenset('di')
    UNSIGNED_INT = frozenset('oxXup')
    FLOATING_POINT = frozenset('fFeEaAgG')

    _PACKED_FLOAT = struct.Struct('<f')

    @classmethod
    def from_string(cls, format_specifier: str):
        """Creates a FormatSpec from a str with a single format specifier."""
        match = cls.FORMAT_SPEC.fullmatch(format_specifier)

        if not match:
            raise ValueError(
                '{!r} is not a valid single format specifier'.format(
                    format_specifier
                )
            )

        return cls(match)

    def __init__(self, re_match: Match):
        """Constructs a FormatSpec from an re.Match object for FORMAT_SPEC."""
        self.match = re_match
        self.specifier: str = self.match.group()

        self.flags: str = self.match.group('flags') or ''
        self.width: str = self.match.group('width') or ''
        self.precision: str = self.match.group('precision') or ''
        self.length: str = self.match.group('length') or ''
        self.type: str = self.match.group('type')

        self.error = None
        if self.type == 'n':
            self.error = 'Unsupported conversion specifier n.'
        elif self.type == '%':
            if self.flags or self.width or self.precision or self.length:
                self.error = (
                    '%% does not support any flags, width, precision,'
                    'or length modifiers.'
                )
        elif self.type in 'csdiup' and '#' in self.flags:
            self.error = (
                '# is only supported with o, x, X, f, F, e, E, a, A, '
                'g, and G specifiers.'
            )
        elif self.type in 'csp' and '0' in self.flags:
            self.error = (
                '0 is only supported with d, i, o, u, x, X, a, A, e, '
                'E, f, F, g, and G specifiers.'
            )
        elif self.type in 'cs' and ('+' in self.flags or ' ' in self.flags):
            self.error = (
                '+ and space are only available for d, i, o, u, x, X,'
                'a, A, e, E, f, F, g, and G specifiers.'
            )
        elif self.type == 'c':
            if self.precision != '':
                self.error = 'Precision is not supported for specifier c.'
        elif self.type == 'p':
            if self.length != '':
                self.error = 'p does not support any length modifiers.'
            elif self.precision != '':
                self.error = 'p does not support precision modifiers.'

        # If we are going to add additional characters to the output, we add to
        # width_bias to ensure user-provided widths are reduced by that amount.
        self._width_bias = 0
        # Some of our machinery requires that we maintain a minimum precision
        # width to ensure a certain amount of digits gets printed. This
        # increases the user-provided precision in these cases if it was not
        # enough.
        self._minimum_precision = 0
        # Python's handling of %#o is non-standard and prepends a 0o
        # instead of single 0.
        if self.type == 'o' and '#' in self.flags:
            self._width_bias = 1
        # Python does not support %p natively.
        if self.type == 'p':
            self._width_bias = 2
            self._minimum_precision = 8

        # If we have a concrete width, we reduce it by any width bias.
        # Otherwise, we either have no width or width is *, where the decoding
        # logic will handle the width bias.
        parsed_width = int(self.width.replace('*', '') or '0')
        if parsed_width > self._width_bias:
            self.width = f'{parsed_width - self._width_bias}'

        # Python %-operator does not support `.` without a
        # trailing number. `.` is defined to be equivalent to `.0`.
        if self.precision == '.':
            self.precision = '.0'

        # If we have a concrete precision that is not *, we check that it is at
        # least minimum precision. If it is *, other parts of decoding will
        # ensure the minimum is upheld.
        if (
            self.precision != '.*'
            and int(self.precision.replace('.', '') or '0')
            < self._minimum_precision
        ):
            self.precision = f'.{self._minimum_precision}'

        # The Python %-format machinery never requires the length
        # modifier to work correctly, and it doesn't support all of the
        # C99 length format specifiers anyway. We remove it from the
        # python-compaitble format string.
        self.compatible = ''.join(
            [
                '%',
                self.flags,
                self.width,
                self.precision,
                self._REMAP_TYPE.get(self.type, self.type),
            ]
        )

    def decode(self, encoded_arg: bytes) -> 'DecodedArg':
        """Decodes the provided data according to this format specifier."""
        if self.error is not None:
            return DecodedArg(
                self, None, b'', DecodedArg.DECODE_ERROR, self.error
            )

        width = None
        if self.width == '*':
            width = FormatSpec.from_string('%d').decode(encoded_arg)
            encoded_arg = encoded_arg[len(width.raw_data) :]

        precision = None
        if self.precision == '.*':
            precision = FormatSpec.from_string('%d').decode(encoded_arg)
            encoded_arg = encoded_arg[len(precision.raw_data) :]

        if self.type == '%':
            return DecodedArg(
                self, (), b''
            )  # Use () as the value for % formatting.

        if self.type == 's':
            return self._merge_decoded_args(
                width, precision, self._decode_string(encoded_arg)
            )

        if self.type == 'c':
            return self._merge_decoded_args(
                width, precision, self._decode_char(encoded_arg)
            )

        if self.type in self.SIGNED_INT:
            return self._merge_decoded_args(
                width, precision, self._decode_signed_integer(encoded_arg)
            )

        if self.type in self.UNSIGNED_INT:
            return self._merge_decoded_args(
                width, precision, self._decode_unsigned_integer(encoded_arg)
            )

        if self.type in self.FLOATING_POINT:
            return self._merge_decoded_args(
                width, precision, self._decode_float(encoded_arg)
            )

        # Should be unreachable.
        assert False, f'Unhandled format specifier: {self.type}'

    def text_float_safe_compatible(self) -> str:
        return ''.join(
            [
                '%',
                self.flags.replace('0', ' '),
                self.width,
                self.precision,
                self._REMAP_TYPE.get(self.type, self.type),
            ]
        )

    def _merge_decoded_args(
        self,
        width: Optional['DecodedArg'],
        precision: Optional['DecodedArg'],
        main: 'DecodedArg',
    ) -> 'DecodedArg':
        def merge_optional_str(*args: Optional[str]) -> Optional[str]:
            return ' '.join(a for a in args if a) or None

        if width is not None and precision is not None:
            return DecodedArg(
                main.specifier,
                (
                    width.value - self._width_bias,
                    max(precision.value, self._minimum_precision),
                    main.value,
                ),
                width.raw_data + precision.raw_data + main.raw_data,
                width.status | precision.status | main.status,
                merge_optional_str(width.error, precision.error, main.error),
            )

        if width is not None:
            return DecodedArg(
                main.specifier,
                (width.value - self._width_bias, main.value),
                width.raw_data + main.raw_data,
                width.status | main.status,
                merge_optional_str(width.error, main.error),
            )

        if precision is not None:
            return DecodedArg(
                main.specifier,
                (max(precision.value, self._minimum_precision), main.value),
                precision.raw_data + main.raw_data,
                precision.status | main.status,
                merge_optional_str(precision.error, main.error),
            )

        return main

    def _decode_signed_integer(
        self,
        encoded: bytes,
    ) -> 'DecodedArg':
        """Decodes a signed variable-length integer."""
        if not encoded:
            return DecodedArg.missing(self)

        count = 0
        result = 0
        shift = 0

        for byte in encoded:
            count += 1
            result |= (byte & 0x7F) << shift

            if not byte & 0x80:
                return DecodedArg(
                    self,
                    zigzag_decode(result),
                    encoded[:count],
                    DecodedArg.OK,
                )

            shift += 7
            if shift >= 64:
                break

        return DecodedArg(
            self,
            None,
            encoded[:count],
            DecodedArg.DECODE_ERROR,
            'Unterminated variable-length integer',
        )

    def _decode_unsigned_integer(self, encoded: bytes) -> 'DecodedArg':
        """Decodes an unsigned variable-length integer."""
        arg = self._decode_signed_integer(encoded)
        # Since ZigZag encoding is used, unsigned integers must be masked off to
        # their original bit length.
        if arg.value is not None:
            arg.value &= (1 << self.size_bits()) - 1

        return arg

    def _decode_float(self, encoded: bytes) -> 'DecodedArg':
        if len(encoded) < 4:
            return DecodedArg.missing(self)

        return DecodedArg(
            self, self._PACKED_FLOAT.unpack_from(encoded)[0], encoded[:4]
        )

    def _decode_string(self, encoded: bytes) -> 'DecodedArg':
        """Reads a unicode string from the encoded data."""
        if not encoded:
            return DecodedArg.missing(self)

        size_and_status = encoded[0]
        status = DecodedArg.OK

        if size_and_status & 0x80:
            status |= DecodedArg.TRUNCATED
            size_and_status &= 0x7F

        raw_data = encoded[0 : size_and_status + 1]
        data = raw_data[1:]

        if len(data) < size_and_status:
            status |= DecodedArg.DECODE_ERROR

        try:
            decoded = data.decode()
        except UnicodeDecodeError as err:
            return DecodedArg(
                self,
                repr(bytes(data)).lstrip('b'),
                raw_data,
                status | DecodedArg.DECODE_ERROR,
                err,
            )

        return DecodedArg(self, decoded, raw_data, status)

    def _decode_char(self, encoded: bytes) -> 'DecodedArg':
        """Reads an integer from the data, then converts it to a string."""
        arg = self._decode_signed_integer(encoded)

        if arg.ok():
            try:
                arg.value = chr(arg.value)
            except (OverflowError, ValueError) as err:
                arg.error = err
                arg.status |= DecodedArg.DECODE_ERROR

        return arg

    def size_bits(self) -> int:
        """Size of the argument in bits; 0 for strings."""
        if self.type == 's':
            return 0

        # TODO(hepler): 64-bit targets likely have 64-bit l, j, z, and t.
        return 64 if self.length in ['ll', 'j'] else 32

    def __str__(self) -> str:
        return self.specifier


class DecodedArg:
    """Represents a decoded argument that is ready to be formatted."""

    # Status flags for a decoded argument. These values should match the
    # DecodingStatus enum in pw_tokenizer/internal/decode.h.
    OK = 0  # decoding was successful
    MISSING = 1  # the argument was not present in the data
    TRUNCATED = 2  # the argument was truncated during encoding
    DECODE_ERROR = 4  # an error occurred while decoding the argument
    SKIPPED = 8  # argument was skipped due to a previous error

    @classmethod
    def missing(cls, specifier: FormatSpec):
        return cls(specifier, None, b'', cls.MISSING)

    def __init__(
        self,
        specifier: FormatSpec,
        value,
        raw_data: bytes,
        status: int = OK,
        error=None,
    ):
        self.specifier = specifier  # FormatSpec (e.g. to represent "%0.2f")
        self.value = value  # the decoded value, or None if decoding failed
        self.raw_data = bytes(
            raw_data
        )  # the exact bytes used to decode this arg
        self._status = status
        self.error = error

    def ok(self) -> bool:
        """The argument was decoded without errors."""
        return self.status == self.OK or self.status == self.TRUNCATED

    @property
    def status(self) -> int:
        return self._status

    @status.setter
    def status(self, status: int):
        # The %% specifier is always OK and should always be printed normally.
        self._status = status if self.specifier.type != '%' else self.OK

    def format(self) -> str:
        """Returns formatted version of this argument, with error handling."""
        if self.status == self.TRUNCATED:
            return self.specifier.compatible % (self.value + '[...]')

        if self.ok():
            # Check if we are effectively .0{diuoxX} with a 0 value (this
            # includes .* with (0, 0)). C standard says a value of 0 with 0
            # precision produces an empty string.
            is_integer_specifier_type = self.specifier.type in 'diuoxX'
            is_simple_0_precision_with_0_value = self.value == 0 and (
                self.specifier.precision == '.0'
                or self.specifier.precision == '.'
            )
            is_star_0_precision_with_0_value = (
                self.value == (0, 0) and self.specifier.precision == '.*'
            )
            if is_integer_specifier_type and (
                is_simple_0_precision_with_0_value
                or is_star_0_precision_with_0_value
            ):
                return ''

            try:
                # Python has a nonstandard alternative octal form.
                if self.specifier.type == 'o' and '#' in self.specifier.flags:
                    return self._format_alternative_octal()

                # Python doesn't pad zeros correctly for inf/nan.
                if self.specifier.type in FormatSpec.FLOATING_POINT and (
                    self.value == math.inf
                    or self.value == -math.inf
                    or self.value == math.nan
                ):
                    return self._format_text_float()

                # Python doesn't have a native pointer formatter.
                if self.specifier.type == 'p':
                    return self._format_pointer()

                return self.specifier.compatible % self.value
            except (OverflowError, TypeError, ValueError) as err:
                self._status |= self.DECODE_ERROR
                self.error = err

        if self.status & self.SKIPPED:
            message = '{} SKIPPED'.format(self.specifier)
        elif self.status == self.MISSING:
            message = '{} MISSING'.format(self.specifier)
        elif self.status & self.DECODE_ERROR:
            message = '{} ERROR'.format(self.specifier)
        else:
            raise AssertionError(
                'Unhandled DecodedArg status {:x}!'.format(self.status)
            )

        if self.value is None or not str(self.value):
            return '<[{}]>'.format(message)

        return '<[{} ({})]>'.format(message, self.value)

    def _format_alternative_octal(self) -> str:
        """Formats an alternative octal specifier.

        This potentially throws OverflowError, TypeError, or ValueError.
        """
        compatible_specifier = self.specifier.compatible.replace('#', '')
        result = compatible_specifier % self.value

        # Find index of the first non-space, non-plus, and non-zero
        # character. If we cannot find anything, we will simply
        # prepend a 0 to the formatted string.
        counter = 0
        for i, value in enumerate(result):
            if value not in ' +0':
                counter = i
                break
        return result[:counter] + '0' + result[counter:]

    def _format_text_float(self) -> str:
        """Formats a float specifier with txt value (e.g. NAN, INF).

        This potentially throws OverflowError, TypeError, or ValueError.
        """
        return self.specifier.text_float_safe_compatible() % self.value

    def _format_pointer(self) -> str:
        """Formats a pointer specifier.

        This potentially throws OverflowError, TypeError, or ValueError.
        """
        result = self.specifier.compatible % self.value

        # Find index of the first non-space, non-plus, and non-zero
        # character (unless we hit the first of the 8 required hex
        # digits).
        counter = 0
        for i, value in enumerate(result[:-7]):
            if value not in ' +0' or i == len(result) - 8:
                counter = i
                break

        # Insert the pointer 0x prefix in after the leading `+`,
        # space, or `0`
        return result[:counter] + '0x' + result[counter:]

    def __str__(self) -> str:
        return self.format()

    def __repr__(self) -> str:
        return f'DecodedArg({self})'


def parse_format_specifiers(format_string: str) -> Iterable[FormatSpec]:
    for spec in FormatSpec.FORMAT_SPEC.finditer(format_string):
        yield FormatSpec(spec)


class FormattedString(NamedTuple):
    value: str
    args: Sequence[DecodedArg]
    remaining: bytes

    def ok(self) -> bool:
        """Arg data decoded successfully and all expected args were found."""
        return all(arg.ok() for arg in self.args) and not self.remaining

    def score(self, date_removed: Optional[datetime] = None) -> tuple:
        """Returns a key for sorting by how successful a decode was.

        Decoded strings are sorted by whether they

          1. decoded all bytes for all arguments without errors,
          2. decoded all data,
          3. have the fewest decoding errors,
          4. decoded the most arguments successfully, or
          5. have the most recent removal date, if they were removed.

        This must match the collision resolution logic in detokenize.cc.

        To format a list of FormattedStrings from most to least successful,
        use sort(key=FormattedString.score, reverse=True).
        """
        return (
            self.ok(),  # decocoded all data and all expected args were found
            not self.remaining,  # decoded all data
            -sum(not arg.ok() for arg in self.args),  # fewest errors
            len(self.args),  # decoded the most arguments
            date_removed or datetime.max,
        )  # most recently present


class FormatString:
    """Represents a printf-style format string."""

    def __init__(self, format_string: str):
        """Parses format specifiers in the format string."""
        self.format_string = format_string
        self.specifiers = tuple(parse_format_specifiers(self.format_string))

        # List of non-specifier string pieces with room for formatted arguments.
        self._segments = self._parse_string_segments()

    def _parse_string_segments(self) -> List:
        """Splits the format string by format specifiers."""
        if not self.specifiers:
            return [self.format_string]

        spec_spans = [spec.match.span() for spec in self.specifiers]

        # Start with the part of the format string up to the first specifier.
        string_pieces = [self.format_string[: spec_spans[0][0]]]

        for (_, end1), (start2, _) in zip(spec_spans[:-1], spec_spans[1:]):
            string_pieces.append(self.format_string[end1:start2])

        # Append the format string segment after the last format specifier.
        string_pieces.append(self.format_string[spec_spans[-1][1] :])

        # Make a list with spots for the replacements between the string pieces.
        segments: List = [None] * (len(string_pieces) + len(self.specifiers))
        segments[::2] = string_pieces

        return segments

    def decode(self, encoded: bytes) -> Tuple[Sequence[DecodedArg], bytes]:
        """Decodes arguments according to the format string.

        Args:
          encoded: bytes; the encoded arguments

        Returns:
          tuple with the decoded arguments and any unparsed data
        """
        decoded_args = []

        fatal_error = False
        index = 0

        for spec in self.specifiers:
            arg = spec.decode(encoded[index:])

            if fatal_error:
                # After an error is encountered, continue to attempt to parse
                # arguments, but mark them all as SKIPPED. If an error occurs,
                # it's impossible to know if subsequent arguments are valid.
                arg.status |= DecodedArg.SKIPPED
            elif not arg.ok():
                fatal_error = True

            decoded_args.append(arg)
            index += len(arg.raw_data)

        return tuple(decoded_args), encoded[index:]

    def format(
        self, encoded_args: bytes, show_errors: bool = False
    ) -> FormattedString:
        """Decodes arguments and formats the string with them.

        Args:
          encoded_args: the arguments to decode and format the string with
          show_errors: if True, an error message is used in place of the %
              conversion specifier when an argument fails to decode

        Returns:
          tuple with the formatted string, decoded arguments, and remaining data
        """
        # Insert formatted arguments in place of each format specifier.
        args, remaining = self.decode(encoded_args)

        if show_errors:
            self._segments[1::2] = (arg.format() for arg in args)
        else:
            self._segments[1::2] = (
                arg.format() if arg.ok() else arg.specifier.specifier
                for arg in args
            )

        return FormattedString(''.join(self._segments), args, remaining)


def decode(
    format_string: str, encoded_arguments: bytes, show_errors: bool = False
) -> str:
    """Decodes arguments and formats them with the provided format string.

    Args:
      format_string: the printf-style format string
      encoded_arguments: encoded arguments with which to format
          format_string; must exclude the 4-byte string token
      show_errors: if True, an error message is used in place of the %
          conversion specifier when an argument fails to decode

    Returns:
      the printf-style formatted string
    """
    return (
        FormatString(format_string).format(encoded_arguments, show_errors).value
    )



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Copies files from CIPD to a specified directory.

By default, Pigweed installs packages from a manifest file to a CIPD
subdirectory as part of environment setup. This script will copy files from this
directory into a specified output directory.

Here's an example of how to use this script:

Let's say you have a package with a static library:

CIPD path: `pigweed/third_party/libsomething`
Files:
  ./libsomething/include/something.h
  ./libsomething/libsomething.a

And this package was referenced in my_project_packages.json, which was provided
as a --cipd-package-file in your bootstrap script.

To copy the static libraryto $PW_PROJECT_ROOT/static_libraries, you'd have an
invocation something like this:

copy_from_cipd --package-name=pigweed/third_party/libsomething \
               --mainfest=$PW_PROJECT_ROOT/tools/my_project_packages.json \
               --file=libsomething/libsomething.a \
               --out=$PW_PROJECT_ROOT/static_libraries
"""

import argparse
import json
import logging
import os
import shutil
import subprocess
import sys
from pathlib import Path

import pw_env_setup.cipd_setup.update

logger = logging.getLogger(__name__)


def parse_args():
    """Parse arguments."""
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        '--verbose', '-v', help='Verbose output', action='store_true'
    )
    parser.add_argument(
        '--manifest',
        required=True,
        type=Path,
        help='Path to CIPD JSON manifest file',
    )
    parser.add_argument(
        '--out-dir',
        type=Path,
        default='.',
        help='Output folder to copy the specified file to',
    )
    parser.add_argument(
        '--package-name', required=True, help='The CIPD package name'
    )
    # TODO(pwbug/334) Support multiple values for --file.
    parser.add_argument(
        '--file',
        required=True,
        type=Path,
        help='Path of the file to copy from the CIPD package. '
        'This is relative to the CIPD package root of the '
        'provided manifest.',
    )
    parser.add_argument(
        '--cipd-package-root',
        type=Path,
        help="Path to the root of the package's install "
        'directory. This is usually at '
        'PW_{manifest name}_CIPD_INSTALL_DIR',
    )
    return parser.parse_args()


def check_version(manifest, cipd_path, package_name):
    base_package_name = os.path.basename(package_name)
    instance_id_path = os.path.join(
        cipd_path, '.versions', f'{base_package_name}.cipd_version'
    )
    with open(instance_id_path, 'r') as ins:
        instance_id = json.load(ins)['instance_id']

    with open(manifest, 'r') as ins:
        data = json.load(ins)['packages']

    path = None
    expected_version = None
    for entry in data:
        if package_name in entry['path']:
            path = entry['path']
            expected_version = entry['tags'][0]
    if not path:
        raise LookupError(f'failed to find {package_name} entry')

    cmd = ['cipd', 'describe', path, '-version', instance_id]
    output = subprocess.check_output(cmd).decode()
    if expected_version not in output:
        pw_env_setup.cipd_setup.update.update(
            'cipd',
            (manifest,),
            os.environ['PW_CIPD_INSTALL_DIR'],
            os.environ['CIPD_CACHE_DIR'],
        )


def main():
    args = parse_args()

    if args.verbose:
        logger.setLevel(logging.DEBUG)

    # Try to infer CIPD install root from the manifest name.
    if args.cipd_package_root is None:
        file_base_name = args.manifest.stem
        args.cipd_var = 'PW_{}_CIPD_INSTALL_DIR'.format(file_base_name.upper())
        try:
            args.cipd_package_root = os.environ[args.cipd_var]
        except KeyError:
            logger.error(
                "The %s environment variable isn't set. Did you forget to run "
                '`. ./bootstrap.sh`? Is the %s manifest installed to a '
                'different path?',
                args.cipd_var,
                file_base_name,
            )
            sys.exit(1)

    check_version(args.manifest, args.cipd_package_root, args.package_name)

    shutil.copyfile(
        os.path.join(args.cipd_package_root, args.file),
        os.path.join(args.out_dir, args.file),
    )


if __name__ == '__main__':
    logging.basicConfig()
    main()



# Copyright 2019 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Python wrapper that runs a program. For use in GN."""

import argparse
import logging
import os
import re
import shlex
import subprocess
import sys
import pathlib
from typing import Dict, Optional

# Need to be able to run without pw_cli installed in the virtualenv.
try:
    import pw_cli.log
except ImportError:
    pass

_LOG = logging.getLogger(__name__)


def argument_parser(
    parser: Optional[argparse.ArgumentParser] = None,
) -> argparse.ArgumentParser:
    """Registers the script's arguments on an argument parser."""

    if parser is None:
        parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument(
        '--args-file',
        type=argparse.FileType('r'),
        help='File containing extra positional arguments to the program',
    )
    parser.add_argument(
        '--capture-output',
        action='store_true',
        help='Hide output from the program unless it fails',
    )
    parser.add_argument(
        '-e',
        '--env',
        action='append',
        default=[],
        help='key=value environment pair for the process',
    )
    parser.add_argument(
        '--env-file',
        type=argparse.FileType('r'),
        help='File defining environment variables for the process',
    )
    parser.add_argument(
        '--skip-empty-args',
        action='store_true',
        help='Don\'t run the program if --args-file is empty',
    )
    parser.add_argument(
        '--target',
        help='GN build target that runs the program',
    )
    parser.add_argument(
        '--working-directory',
        type=pathlib.Path,
        help='Directory to execute program in',
    )
    parser.add_argument(
        'command',
        nargs=argparse.REMAINDER,
        help='Program to run with arguments',
    )

    return parser


_ENV_REGEX = re.compile(r'(\w+)(\+)?=(.+)')


def apply_env_var(string: str, env: Dict[str, str]) -> None:
    """Update an environment map with provided a key-value pair.

    Pairs are accepted in two forms:

      KEY=value    sets environment variable "KEY" to "value"
      KEY+=value   appends OS-specific PATH separator and "value" to
                   environment variable "KEY"
    """
    result = _ENV_REGEX.search(string.strip())
    if not result:
        return

    key, append, val = result.groups()
    if append is not None:
        curr = env.get(key)
        val = f'{curr}{os.path.pathsep}{val}' if curr else val

    env[key] = val


def main() -> int:
    """Runs a program specified by command-line arguments."""
    args = argument_parser().parse_args()
    if not args.command or args.command[0] != '--':
        return 1

    env = os.environ.copy()

    # Command starts after the "--".
    command = args.command[1:]
    # command[0] is the invoker.prog from gn and gn will escape
    # the various spaces in the command which means when argparse
    # gets the string argparse believes this as a single argument
    # and cannot correctly break the string into a list that
    # subprocess can handle.  By splitting the first element
    # in the command list, if there is a space, all of the
    # command[0] elements will be made into a list and if not
    # then split won't do everything and the old behavior
    # will continue.
    front_command = command[0].split(' ')
    del command[0]
    command = front_command + command
    extra_kw_args = {}

    if args.args_file is not None:
        empty = True
        for line in args.args_file:
            empty = False
            command.append(line.strip())

        if args.skip_empty_args and empty:
            return 0

    if args.env_file is not None:
        for line in args.env_file:
            apply_env_var(line, env)

    # Apply command-line overrides at a higher priority than the env file.
    for string in args.env:
        apply_env_var(string, env)

    if args.capture_output:
        extra_kw_args['stdout'] = subprocess.PIPE
        extra_kw_args['stderr'] = subprocess.STDOUT

    if args.working_directory:
        extra_kw_args['cwd'] = args.working_directory

    process = subprocess.run(command, env=env, **extra_kw_args)  # type: ignore

    if process.returncode != 0 and args.capture_output:
        _LOG.error('')
        _LOG.error(
            'Command failed with exit code %d in GN build.', process.returncode
        )
        _LOG.error('')
        _LOG.error('Build target:')
        _LOG.error('')
        _LOG.error('  %s', args.target)
        _LOG.error('')
        _LOG.error('Full command:')
        _LOG.error('')
        _LOG.error('  %s', ' '.join(shlex.quote(arg) for arg in command))
        _LOG.error('')
        _LOG.error('Process output:')
        print(flush=True)
        sys.stdout.buffer.write(process.stdout)
        print(flush=True)
        _LOG.error('')

    return process.returncode


if __name__ == '__main__':
    # If pw_cli is not yet installed in the virtualenv just skip it.
    if 'pw_cli' in globals():
        pw_cli.log.install()
    sys.exit(main())



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_protobuf_codegen_protos/codegen_options.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_protobuf_codegen_protos/codegen_options.proto',
  package='pw.protobuf',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n0pw_protobuf_codegen_protos/codegen_options.proto\x12\x0bpw.protobuf\"t\n\x0e\x43odegenOptions\x12\x11\n\tmax_count\x18\x01 \x01(\x05\x12\x13\n\x0b\x66ixed_count\x18\x02 \x01(\x08\x12\x10\n\x08max_size\x18\x03 \x01(\x05\x12\x12\n\nfixed_size\x18\x04 \x01(\x08\x12\x14\n\x0cuse_callback\x18\x05 \x01(\x08\x62\x06proto3'
)




_CODEGENOPTIONS = _descriptor.Descriptor(
  name='CodegenOptions',
  full_name='pw.protobuf.CodegenOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='max_count', full_name='pw.protobuf.CodegenOptions.max_count', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fixed_count', full_name='pw.protobuf.CodegenOptions.fixed_count', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_size', full_name='pw.protobuf.CodegenOptions.max_size', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fixed_size', full_name='pw.protobuf.CodegenOptions.fixed_size', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='use_callback', full_name='pw.protobuf.CodegenOptions.use_callback', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=65,
  serialized_end=181,
)

DESCRIPTOR.message_types_by_name['CodegenOptions'] = _CODEGENOPTIONS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CodegenOptions = _reflection.GeneratedProtocolMessageType('CodegenOptions', (_message.Message,), {
  'DESCRIPTOR' : _CODEGENOPTIONS,
  '__module__' : 'pw_protobuf_codegen_protos.codegen_options_pb2'
  # @@protoc_insertion_point(class_scope:pw.protobuf.CodegenOptions)
  })
_sym_db.RegisterMessage(CodegenOptions)


# @@protoc_insertion_point(module_scope)
